"""Transaction-aware Celery task handling.

Inspired by Warehouse project https://raw.githubusercontent.com/pypa/warehouse/master/warehouse/celery.py
"""
# Standard Library
import logging

# Pyramid
import transaction
import venusian
from pyramid.request import apply_request_extensions
from pyramid.scripting import _make_request
from transaction import TransactionManager

# Celery
from celery import Task

# Websauna
from websauna.system.http import Request
from websauna.system.model.retry import retryable
from websauna.system.task.celery import get_celery


logger = logging.getLogger(__name__)


class WebsaunaTask(Task):
    """A task that can clean up its transaction at the end."""

    def get_request(self, **options) -> Request:
        """Get the current HTTPRequest interface associated with the task.

        This is not a real HTTP request - Celery is not connected to HTTP interface. Instead, a faux request is generated by :py:class:`websauna.system.task.celeryloader.WebsaunaLoader`.
        """
        # This is set by on_task_init in loader
        request = getattr(self.request, "request", None)
        if request:
            return request

        return request

    def is_eager(self) -> bool:
        return self.request.is_eager

    def on_failure(self, exc, task_id, args, kwargs, einfo):
        """What happens if a task raises exception."""
        # Should be logged by properly configured Celery itself
        # logger.error("Celery task failure %s, args %s, kwargs %s", task_id, args, kwargs)
        pass

    def after_return(self, status, retval, task_id, args, kwargs, einfo):
        """Clean up transaction after task run."""
        if not self.request.is_eager:
            # Close the request when task completes
            request = self.get_request()

            # Make sure tasks don't leave transaction open e.g. in the case of exception
            if status == "FAILURE":
                logger.debug("Closing request task %s, status %s", self, status)
                tm = request.transaction_manager
                txn = tm._txn
                if txn:
                    txn.abort()
            else:
                logger.debug("Finished request task %s, status %s", self, status)
                # This will terminate dbsession, as set in create_transaction_manager_aware_dbsession

            # Call add-on hooks
            from websauna.system.task.events import TaskFinished  # Avoid circular imports
            request.registry.notify(TaskFinished(request, self))

            request._process_finished_callbacks()


class ScheduleOnCommitTask(WebsaunaTask):
    """A Celery task that does not get scheduled to execution until the current transaction commits.

    This is a :py:class:`celery.app.task.Task` based class to be used with :py:meth:`celery.Celery.task` function decorator.

    The created task only executes through ``apply_async`` if the web transaction successfully commits and only after transaction successfully commits. Thus, it is safe to pass ids to any database objects for the task and expect the task to be able to read them.
    """

    def make_faux_request(self, transaction_manager=None):
        """In the case we can't use real request object, make a new request from registry given to Celery."""
        # Real request has been already committed in this point,
        # so create a faux request to satisfy the presence of dbsession et. al.
        registry = self.app.registry

        request = _make_request("/", registry)

        # Make sure we have a transaction manager
        if not transaction_manager:
            transaction_manager = transaction.manager

        request.tm = transaction_manager

        apply_request_extensions(request)
        return request

    def get_transaction_manager(self, **options) -> TransactionManager:
        """Get the transaction manager we are bound to."""

        tm = options.get("tm")

        if not tm:
            raise RuntimeError("You need to explicitly pass transaction manager as 'tm' task option to ScheduleOnCommitTask. Task keyword arguments are are: {}".format(options))

        return tm

    def exec_eager(self, *args, **kwargs):
        """Run transaction aware task in eager mode."""

        # We are run in a post-commit hook, so there is no transaction manager available
        tm = TransactionManager()

        # Do not attempt any transaction retries in eager mode
        tm.retry_attempt_count = 1

        self.request.update(request=self.make_faux_request(transaction_manager=tm))
        return self.run(*args, **kwargs)

    def __call__(self, *args, **kwargs):
        """Call Celery task and insert request argument.

        Wrap Celery task call for better exception handling.

        Celery itself does very bad job in logging exceptions. So LET'S "#€!"€! STOP SILENTLY SWALLOWING THEM. This is for eager.
        """

        if self.request.is_eager:
            return self.exec_eager(*args, **kwargs)

        try:
            underlying = super().__call__
            return underlying(*args, **kwargs)
        except Exception as e:
            logger.error("Celery task raised an exception %s", e)
            logger.exception(e)
            raise
        finally:
            # TODO? Do we need closer?
            # pyramid_env["closer"]()
            pass

    def apply_async_on_commit(self, *args, **kwargs):
        """Schedule a task from web process."""

        tm = self.get_transaction_manager(**kwargs)
        kwargs.pop("tm")

        logger.debug("Setting after commit hook tm is %s", tm)

        # This will break things that expect to get an AsyncResult because
        # we're no longer going to be returning an async result from this when
        # called from within a request, response cycle. Ideally we shouldn't be
        # waiting for responses in a request/response cycle anyways though.
        tm.get().addAfterCommitHook(
            self._after_commit_hook,
            args=args,
            kws=kwargs,
        )

    def apply_async_instant(self, *args, **options):
        """Schedule async task from a beat process or another task.

        Doesnt' wait a commit to finish to schedule a task.
        You usually don't want to cal this directly, but call `apply_async` from Celery normal interface.
        """
        return super().apply_async(*args, **options)

    def apply_async(self, *args, **options):

        if "producer" in options:
            # This comes from Celery beat process.
            # Celery beat doesn't know about transaction lifecycles.
            # Instantly schedule the task.
            return self.apply_async_instant(*args, **options)
        else:
            # This call comes from inside a web process and
            # we only want to make the task run on commit
            return self.apply_async_on_commit(*args, **options)

    def _after_commit_hook(self, success, *args, **kwargs):
        """When HTTP request terminates and the transaction is committed, actually submit the task to Celery."""

        logger.debug("Calling after commit hook")

        if success:
            result = super().apply_async(*args, **kwargs)

            logger.debug("Commit hook resulted to a Celery task %s", result)


class RetryableTransactionTask(ScheduleOnCommitTask):
    """Celery task that commits all the work at the end of the task using transaction manager commit.

    A base class to be used with :py:meth:`celery.Celery.task` function decorator. Automatically commits all the work at the end of the task.

    In the case of transaction conflict, the task will rerun based on :func:`pyramid_tm.tm_tween_factory` attempt rules.
    """

    abstract = True

    def exec_eager(self, *args, **kwargs):
        """Run transaction aware task in eager mode."""

        # We are run in a post-commit hook, so there is no transaction manager available
        tm = TransactionManager()

        # Do not attempt any transaction retries in eager mode
        tm.retry_attempt_count = 1

        self.request.update(request=self.make_faux_request(transaction_manager=tm))

        with tm:
            # This doesn't do transaction retry attempts, but should be good enough for eager
            return self.run(*args, **kwargs)

    def __call__(self, *args, **kwargs):

        if self.request.is_eager:
            return self.exec_eager(*args, **kwargs)

        request = self.get_request()
        task = self

        try:
            # Celery 4.0+
            # Here we call directly run, because celery.app.Task.__call__ messes with thread locals clearing the task context. Thus, the second transaction attempt would file Task.get_request() == None

            @retryable(tm=request.tm)
            def handler(request: Request):
                if task.__self__ is not None:
                    return task.run(task.__self__, *args, **kwargs)
                return task.run(*args, **kwargs)

            result = handler(request)
        finally:
            # TODO: Do we need closer?
            # pyramid_env["closer"]()
            pass

        return result


class TaskProxy:
    """Late-bind Celery tasks to decorated functions.

    Normally ``celery.task()`` binds everything during import time. But we want to avoid this, as we don't want to deal with any configuration during import time.

    We wrap a decorated function with this proxy. Then we forward all the calls to Celery Task object after it has been bound during the end of configuration.
    """

    def __init__(self, original_func):
        self.original_func = original_func
        self.celery_task = None

        # Venusian setup
        self.__venusian_callbacks__ = None
        self.__name__ = self.original_func.__name__

    def __str__(self):
        return "TaskProxy for {} bound to task {}".format(self.original_func, self.celery_task)

    def __repr__(self):
        return self.__str__()

    def __call__(self, *args, **kwargs):
        raise RuntimeError("Tasked functions should not be directly called. Instead use apply_async() and other Celery task functions to initiate them")

    def bind_celery_task(self, celery_task: Task):
        assert isinstance(celery_task, Task)
        self.celery_task = celery_task

    def __getattr__(self, item):
        """Resolve all method calls to the underlying task."""
        if not self.celery_task:
            raise RuntimeError("Celery task creation failed. Did config.scan() do a sweep on {}? TaskProxy tried to look up attribute: {}".format(self.original_func, item))

        return getattr(self.celery_task, item)


def task(*args, **kwargs):
    """Configuration compatible task decorator.

    Tasks are picked up by :py:meth:`pyramid.config.Configurator.scan` run on the module, not during import time.
    Otherwise we mimic the behavior of :py:meth:`celery.Celery.task`.

    :param args: Passed to Celery task decorator
    :param kwargs: Passed to Celery task decorator
    """

    def _inner(func):
        "The class decorator example"

        proxy = TaskProxy(func)

        def register(scanner, name, task_proxy):
            config = scanner.config
            registry = config.registry
            celery = get_celery(registry)
            celery_task = celery.task(task_proxy.original_func, *args, **kwargs)
            proxy.bind_celery_task(celery_task)

        venusian.attach(proxy, register, category='celery')
        return proxy

    return _inner



"""Deform's resource_registry mechanism to allow widgets to include CSS and JS on the page."""
# Pyramid
from deform import Form
from deform.widget import ResourceRegistry as _ResourceRegistry

# Websauna
from websauna.system.form.interfaces import IFormResources
from websauna.system.http import Request


class ResourceRegistry(_ResourceRegistry):
    """A resource registry that maintains dynamically included CSS and JS files for the page.

     * During view processing, widgets and such can request that their helper JS and CSS is included on the page rendering

     * :py:class:`websauna.system.core.render.OnDemandResourceRenderer` maintains a state whether JS should go the end of ``<body>`` or ``<head>``

    The best practice is to place all JavxaScript files before the closing ``</body>`` tag, so that they do not block the the page rendering and decrease the page response time. However, out there exist a lot of badly written JavaScript code which does not honour this - they have ``<script>`` tags in the middle of ``<body>`` and even worse assume you have loaded dependencies, like jQuery, in the ``<head>``. Namely, Deform 2 form library does this.

    ``BodyRelocatableResourceRegistry`` contains the logic to determine if ``<script>`` tags should go to the end of the page (preferred) or to ``<head>`` (in case some javascript makes assumptions of this). If any widget requires resources assume JS must go to head.

    See also :py:attr`deform.widget.default_resources` which includes the list of default resources activated for any form on the page.

    More information

    * https://developers.google.com/speed/docs/insights/BlockingJS

    * https://developers.google.com/web/fundamentals/performance/critical-rendering-path/adding-interactivity-with-javascript#parser-blocking-vs-asynchronous-javascript
    """

    def __init__(self, request: Request):
        # Load default resoucres from configuration
        form_resources = request.registry.getUtility(IFormResources)
        self.registry = form_resources.get_default_resources().copy()

    def pull_in_resources(self, request: Request, form: Form):
        """Add resources CSS and JS resources from Deform form to a Websauna rendering loop."""
        on_demand_resource_renderer = request.on_demand_resource_renderer

        #: TODO: Support for absolute URLs and such

        for css_url in self.get_widget_css_urls(request, form):
            on_demand_resource_renderer.request_resource("css", css_url)

        for js_url in self.get_widget_js_urls(request, form):
            on_demand_resource_renderer.request_resource("js", js_url, js_requires_head=True)

    def get_widget_js_urls(self, request: Request, form: Form):
        """Generate JS and CSS tags for a widget.

        For demo purposes only - you might have something specific to your application here.

        See http://docs.pylonsproject.org/projects/deform/en/latest/widget.html#the-high-level-deform-field-get-widget-resources-method
        """
        resources = form.get_widget_resources()
        js_resources = resources['js']
        js_links = [request.static_url(r) for r in js_resources]
        return js_links

    def get_widget_css_urls(self, request: Request, form: Form):
        """Generate JS and CSS tags for a widget.

        For demo purposes only - you might have something specific to your application here.

        See http://docs.pylonsproject.org/projects/deform/en/latest/widget.html#the-high-level-deform-field-get-widget-resources-method
        """
        resources = form.get_widget_resources()
        css_resources = resources['css']
        css_links = [request.static_url(r) for r in css_resources]
        return css_links



# Pyramid
from pyramid.httpexceptions import HTTPOk
from pyramid.view import view_config

# Websauna
from websauna.system.core.viewconfig import view_overrides


class ParentResource:
    id = "parent"


class ChildResource(ParentResource):
    id = "child"


class ChildResource2(ParentResource):
    id = "child2"


class GrandChildResource(ChildResource):
    id = "grand_child"


class GrandGrandChildResource(GrandChildResource):
    id = "grand_grand_child"


class ParentView:

    def __init__(self, context, request):
        self.context = context
        self.request = request

    @view_config(name="edit", context=ParentResource)
    def traversing_test(self):
        return HTTPOk("Editing: {}".format(self.context.id))


@view_overrides(context=ChildResource)
class ChildView(ParentView):
    pass


@view_overrides(context=GrandChildResource)
class GrandChildView(ParentView):
    pass


@view_overrides(context=GrandGrandChildResource)
class GrandChildView(GrandChildView):
    pass


@view_overrides(context=ChildResource2)
class ChildViewWithOtherViewConfigs(ParentView):

    @view_config(name="show", context=ChildResource2)
    def traversing_test(self):
        return HTTPOk("Showing: {}".format(self.context.id))


class Root:
    """Traversing root."""

    def __init__(self, request):
        pass

    def __getitem__(self, name):
        if name == "parent":
            return ParentResource()
        elif name == "child":
            return ChildResource()
        elif name == "child2":
            return ChildResource2()
        elif name == "grand_child":
            return GrandChildResource()
        elif name == "grand_grand_child":
            return GrandGrandChildResource()


class ParentRouteView:

    def __init__(self, request):
        pass

    @view_config(route_name="parent_hello", renderer="parent.html")
    def traversing_test(self):
        return {"class": self.__class__.__name__}


@view_overrides(route_name="child_hello", renderer="child.html")
class ChildRouteView(ParentRouteView):
    pass



"""Localhost tunneling utilities for running callback tests against external services."""

# Standard Library
import logging
import subprocess
import tempfile
import time
from distutils.spawn import find_executable


logger = logging.getLogger(__name__)


class NgrokTunnel:
    """Build a tunnel to the test webserver using Ngrok service, exposing the test server to the world.

    When you are testing against third party services which call back your site, you need to make sure the test web server can get called. Usually the third party sites allow configuration of fixed callback locations only. Thus, for the testing, on the third party site, we define a fixed callback pointing to ngrok.io location. Then we use ngrok API to route this location to our test webserver at the beginning of the test run.

    Example::

        @pytest.fixture(scope='module')
        def ngrok_tunnel(request:FixtureRequest, ini_settings:dict, web_server:str) -> NgrokTunnel:
            '''A py.test fixture to build tunnel to test SericaPay payment callbacks.

            The SericaPay has a preconfigured test account with a callback pointing to a Ngrok tunnel entry point. When the functional test suite runs SericaPay tests, it builds the tunnel localhost<->ngrok endpoint at the beginning of the test module executing.
            '''

            # Where is our test web server running
            port = urlparse(web_server).port

            # Construct Ngrok tunnel wrapper
            tunnel = NgrokTunnel(port, ini_settings["trees.ngrok_auth_token"], subdomain="trees")

            request.addfinalizer(tunnel.stop)

            tunnel.start()

            return tunnel


        @pytest.mark.serica
        def test_serica_payment(web_server, browser, dbsession, init, ngrok_tunnel):
            '''Do a Serica payment and see it works.'''

    """

    def __init__(self, port: int, auth_token: str, subdomain: str):
        """Build ngrok HTTP tunnel

        Assume ngrok version 2.x. A temporary ngrok configuration file is created where the ngrok token is stored for running the command.

        :param port: localhost port forwarded through tunnel

        :param auth_token: Ngrok auth token to use

        :param subdomain: Use this subdomain for the tunnelling
        """
        assert find_executable("ngrok"), "ngrok command must be installed, see https://ngrok.com/"
        self.port = port
        self.auth_token = auth_token
        self.subdomain = subdomain
        self.ngrok = None
        self.config_file_handle, self.config_file = tempfile.mkstemp()

    def start(self, ngrok_die_check_delay=2.0):
        """Starts the thread on the background and blocks until we get a tunnel URL.

        :param ngrok_die_check_delay: Enough to get the error from ngrok service "Your account is limited to one tunnel"

        :return: the tunnel URL which is now publicly open for your localhost port
        """

        logger.debug("Starting ngrok tunnel %s for port %d", self.subdomain, self.port)

        try:
            subprocess.check_output(["ngrok", "authtoken", self.auth_token, "--config", self.config_file])
        except subprocess.CalledProcessError as e:
            raise AssertionError("Configuring ngrok failed: {}".format(e.output)) from e

        # TODO: capture ngrok output in the case of an error.... no actual ngrok failure reason is not printed
        self.ngrok = subprocess.Popen(["ngrok", "http", str(self.port), "--subdomain", self.subdomain, "--config", self.config_file], stdout=subprocess.DEVNULL)

        # See that we don't instantly die, wait ngrok API to throw us an error
        time.sleep(ngrok_die_check_delay)
        assert self.ngrok.poll() is None, "ngrok terminated abrutly"
        url = "https://{}.ngrok.com".format(self.subdomain)
        return url

    def stop(self):
        """Tell ngrok to tear down the tunnel.

        Stop the background tunneling process.
        """
        if self.ngrok and self.ngrok.returncode is None:
            self.ngrok.terminate()



# Pyramid
import colander
import deform


@colander.deferred
def deferred_csrf_value(node, kw):
    return kw['request'].session.get_csrf_token()


class CSRFSchema(colander.Schema):
    """Schema base class which generates CSRF token.

    Example:

    .. code-block:: python

      from websauna.system.form.schema import CSRFSchema
      import colander

      class MySchema(CSRFSchema):
          my_value = colander.SchemaNode(colander.String())

      And in your application code, *bind* the schema, passing the request as a keyword argument:

      .. code-block:: python

        def aview(request):
            schema = MySchema().bind(request=request)

    The token is automatically then verified by Pyramid CSRF view deriver.

    Original code: https://github.com/Pylons/pyramid_deform/blob/master/pyramid_deform/__init__.py
    """
    csrf_token = colander.SchemaNode(
        colander.String(),
        widget=deform.widget.HiddenWidget(),
        default=deferred_csrf_value,
    )

    csrf_token.dictify_by_default = False


def add_csrf(schema: colander.Schema):
    """Add a hidden csrf_token field on the existing Colander schema."""
    csrf_token = colander.SchemaNode(colander.String(), name="csrf_token", widget=deform.widget.HiddenWidget(), default=deferred_csrf_value)
    csrf_token.dictify_by_default = False
    schema.add(csrf_token)



# Pyramid
import transaction

from flaky import flaky

# Websauna
from websauna.system.user.models import User
from websauna.system.tests.utils import create_logged_in_user
from websauna.utils.slug import uuid_to_slug


GROUP_NAME = "Sample Group"


def get_user(dbsession):
    return dbsession.query(User).first()


def test_add_group(web_server, browser, dbsession, init):
    """Create a new group through admin interface."""

    b = browser
    create_logged_in_user(dbsession, init.config.registry, web_server, browser, admin=True)

    b.find_by_css("#nav-admin").click()

    b.find_by_css("#btn-panel-add-group").click()
    b.fill("name", GROUP_NAME)
    b.fill("description", "Foobar")
    b.find_by_name("add").click()

    assert b.is_text_present("Item added")

    # Check we appear in the list
    b.visit("{}/admin/models/group/listing".format(web_server))

    # The description appears in the listing
    assert b.is_text_present("Foobar")


def test_put_user_to_group(web_server, browser, dbsession, init):
    """Check that we can assign users to groups in admin interface."""

    b = browser

    from websauna.system.user.models import Group

    create_logged_in_user(dbsession, init.config.registry, web_server, browser, admin=True)

    # Create a group where we
    with transaction.manager:
        g = Group(name=GROUP_NAME)
        dbsession.add(g)
        dbsession.flush()
        group_uuid = uuid_to_slug(g.uuid)

    b.find_by_css("#nav-admin").click()
    b.find_by_css("#btn-panel-list-user").click()
    b.find_by_css(".crud-row-1 .btn-crud-listing-edit").click()

    # Check the group checkbox. We could put some more specific classes for controls here.
    b.find_by_css("input[type='checkbox'][value='{}']".format(group_uuid)).click()
    b.find_by_name("save").click()

    assert b.is_text_present("Changes saved")

    # Now we are on Show page of the user, having the new group name visible
    assert b.is_text_present(GROUP_NAME)


@flaky
def test_user_group_choices_preserved_on_validation_error(web_server, init, browser, dbsession):
    """When user edit form validation fails, we should preserve the existing group choices.

    This stresses out hacky implementation of websauna.system.form.colander and deserialization.
    """

    b = browser

    from websauna.system.user.models import Group

    create_logged_in_user(dbsession, init.config.registry, web_server, browser, admin=True)

    # Create a group where we
    with transaction.manager:
        g = Group(name=GROUP_NAME)
        dbsession.add(g)
        u = get_user(dbsession)
        u.groups.append(g)
        dbsession.flush()
        group_uuid = uuid_to_slug(g.uuid)

    b.find_by_css("#nav-admin").click()
    b.find_by_css("#btn-panel-list-user").click()
    b.find_by_css(".crud-row-1 .btn-crud-listing-edit").click()

    # We are in group 2 initially, assert checkbox is checked
    assert b.find_by_css("input[type='checkbox'][value='{}'][checked='True']".format(group_uuid))

    # Do validation error by leaving username empty
    b.fill("username", "")
    b.find_by_name("save").click()
    assert b.is_text_present("There was a problem")

    # Both group checkboxes should be still selected
    with transaction.manager:
        for g in dbsession.query(Group).all():
            assert b.find_by_css("input[type='checkbox'][value='{}'][checked='True']".format(uuid_to_slug(g.uuid)))


def test_remove_user_from_group(web_server, init, browser, dbsession):
    """Remove users from assigned groups in admin."""

    b = browser

    from websauna.system.user.models import Group

    create_logged_in_user(dbsession, init.config.registry, web_server, browser, admin=True)

    # Create a group where we
    with transaction.manager:
        g = Group(name=GROUP_NAME)
        dbsession.add(g)
        u = get_user(dbsession)
        u.groups.append(g)
        dbsession.flush()
        group_uuid = uuid_to_slug(g.uuid)

    b.find_by_css("#nav-admin").click()
    b.find_by_css("#btn-panel-list-user").click()
    b.find_by_css(".crud-row-1 .btn-crud-listing-edit").click()

    # Check the group checkbox. We could put some more specific classes for controls here.
    b.find_by_css("input[type='checkbox'][value='{}']".format(group_uuid)).click()
    b.find_by_name("save").click()

    assert b.is_text_present("Changes saved")

    # After removing we should no longer see the removed group name on user show page
    assert not b.is_text_present(GROUP_NAME)



"""Tests for checking database sanity checks functions correctly."""
# SQLAlchemy
import sqlalchemy
from sqlalchemy import Column
from sqlalchemy import ForeignKey
from sqlalchemy import Integer
from sqlalchemy import String
from sqlalchemy import engine_from_config
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import relationship
from sqlalchemy.orm import sessionmaker

# Websauna
from websauna.system.model.sanitycheck import is_sane_database


def setup_module(self):
    # Quiet log output for the tests
    import logging
    from websauna.system.model.sanitycheck import logger
    logger.setLevel(logging.FATAL)


def gen_test_model():

    Base = declarative_base()

    class SaneTestModel(Base):
        """A sample SQLAlchemy model to demostrate db conflicts. """

        __tablename__ = "sanity_check_test"

        #: Running counter used in foreign key references
        id = Column(Integer, primary_key=True)

    return Base, SaneTestModel


def gen_relation_models():

    Base = declarative_base()

    class RelationTestModel(Base):
        __tablename__ = "sanity_check_test_2"
        id = Column(Integer, primary_key=True)

    class RelationTestModel2(Base):
        __tablename__ = "sanity_check_test_3"
        id = Column(Integer, primary_key=True)

        test_relationship_id = Column(ForeignKey("sanity_check_test_2.id"))
        test_relationship = relationship(RelationTestModel, primaryjoin=test_relationship_id == RelationTestModel.id)

    return Base, RelationTestModel, RelationTestModel2


def gen_declarative():

    Base = declarative_base()

    class DeclarativeTestModel(Base):
        __tablename__ = "sanity_check_test_4"
        id = Column(Integer, primary_key=True)

        @declared_attr
        def _password(self):
            return Column('password', String(256), nullable=False)

        @hybrid_property
        def password(self):
            return self._password

    return Base, DeclarativeTestModel


def test_sanity_pass(ini_settings, dbsession):
    """See database sanity check completes when tables and columns are created."""

    engine = engine_from_config(ini_settings, 'sqlalchemy.')
    conn = engine.connect()
    conn.begin()
    Base, SaneTestModel = gen_test_model()
    Session = sessionmaker(bind=engine)
    session = Session()
    try:
        Base.metadata.drop_all(engine, tables=[SaneTestModel.__table__])
    except sqlalchemy.exc.NoSuchTableError:
        pass

    Base.metadata.create_all(engine, tables=[SaneTestModel.__table__])

    try:
        assert is_sane_database(Base, session) is True
    finally:
        Base.metadata.drop_all(engine)


def test_sanity_table_missing(ini_settings, dbsession):
    """See check fails when there is a missing table"""

    engine = engine_from_config(ini_settings, 'sqlalchemy.')
    conn = engine.connect()
    conn.begin()

    Base, SaneTestModel = gen_test_model()
    Session = sessionmaker(bind=engine)
    session = Session()

    try:
        Base.metadata.drop_all(engine, tables=[SaneTestModel.__table__])
    except sqlalchemy.exc.NoSuchTableError:
        pass

    assert is_sane_database(Base, session) is False


def test_sanity_column_missing(ini_settings, dbsession):
    """See check fails when there is a missing table"""

    engine = engine_from_config(ini_settings, 'sqlalchemy.')
    conn = engine.connect()
    conn.begin()

    Session = sessionmaker(bind=engine)
    session = Session()
    Base, SaneTestModel = gen_test_model()
    try:
        Base.metadata.drop_all(engine, tables=[SaneTestModel.__table__])
    except sqlalchemy.exc.NoSuchTableError:
        pass
    Base.metadata.create_all(engine, tables=[SaneTestModel.__table__])

    # Delete one of the columns
    if engine.dialect.name == "sqlite":
        # SQLite doesn't support alter table
        # TODO: Add some sort of py.test marker
        return
    else:
        engine.execute("ALTER TABLE sanity_check_test DROP COLUMN id")

    assert is_sane_database(Base, session) is False


def test_sanity_pass_relationship(ini_settings, dbsession):
    """See database sanity check understands about relationships and don't deem them as missing column."""

    engine = engine_from_config(ini_settings, 'sqlalchemy.')
    conn = engine.connect()
    conn.begin()

    Session = sessionmaker(bind=engine)
    session = Session()

    Base, RelationTestModel, RelationTestModel2 = gen_relation_models()
    try:
        Base.metadata.drop_all(engine, tables=[RelationTestModel.__table__, RelationTestModel2.__table__])
    except sqlalchemy.exc.NoSuchTableError:
        pass

    Base.metadata.create_all(engine, tables=[RelationTestModel.__table__, RelationTestModel2.__table__])

    try:
        assert is_sane_database(Base, session) is True
    finally:
        Base.metadata.drop_all(engine)


def test_sanity_pass_declarative(ini_settings, dbsession):
    """See database sanity check understands about relationships and don't deem them as missing column."""

    engine = engine_from_config(ini_settings, 'sqlalchemy.')
    conn = engine.connect()
    conn.begin()

    Session = sessionmaker(bind=engine)
    session = Session()

    Base, DeclarativeTestModel = gen_declarative()
    try:
        Base.metadata.drop_all(engine, tables=[DeclarativeTestModel.__table__])
    except sqlalchemy.exc.NoSuchTableError:
        pass

    Base.metadata.create_all(engine, tables=[DeclarativeTestModel.__table__])

    try:
        assert is_sane_database(Base, session) is True
    finally:
        Base.metadata.drop_all(engine)



"""Test ignore session function."""
import pytest

# Websauna
from websauna.system.core import session


test_data = (
    ('/home.html', False),
    ('/admin', False),
    ('/', False),
    ('/static/styles.css', True),
    ('/static/script.js', True),
    ('/static/image.gif', True),
    ('/static/image.jpg', True),
    ('/static/favicon.ico', True),
    ('/static/styles.CSS', True),
    ('/static/script.JS', True),
    ('/static/image.GIF', True),
    ('/static/image.JPG', True),
    ('/static/favicon.ICO', True),
    ('/notebook/static/styles.css', False),
    ('/notebook/static/script.js', False),
    ('/notebook/static/image.gif', False),
    ('/notebook/static/image.jpg', False),
    ('/notebook/static/favicon.ico', False),
)


@pytest.mark.parametrize('url,expected', test_data)
def test_ignore_session(url, expected):
    """Test ignore_session checking."""
    func = session.ignore_session
    assert func(url) is expected



# Pyramid
from zope.interface import Interface


class IModel(Interface):
    """Marker the class which declares SQLAlchemy model.

    This marker interface is used e.g. by ModelAdmin for registry mappings between model and ModelAdmin. We cannot implicitly assume everything is inherited from ``.meta.Base`` because there could be plugin models and such.
    """


class ISQLAlchemySessionFactory(Interface):
    """Provides way to hook in a custom SQLALchemy session creator.

    Use cases include pooling, setting up different session logic for tests, etc.
    """



"""Default HTTP 404 Not Found handling."""

# Standard Library
import logging

# Pyramid
import transaction
from pyramid.renderers import render
from pyramid.response import Response
from pyramid.view import notfound_view_config


logger = logging.getLogger(__name__)


@notfound_view_config()
def notfound(request):
    """Not found view which will log the 404s in the site error log."""

    # Try to extract some more information from request
    user = getattr(request, "user", None)
    if user:
        username = getattr(user, "friendly_name", "<unknown>")
    else:
        username = "<anomymous>"

    # TODO: Maybe make this configurable, default to WARN, configurable as INFO for high volume sites
    logger.info("404 Not Found. user:%s URL:%s referrer:%s", request.url, username, request.referrer)

    # Make sure 404 page does not have any status information, as it is often overlooked special case for caching and we don't want to cache user information
    try:
        request.user = None
    except Exception as exc:
        logger.debug("pyramid_tm 2.0 - this fails: {exc}".format(exc=exc))
        pass

    # The template rendering opens a new transaction which is not rolled back by Pyramid transaction machinery, because we are in a very special view. This tranaction will cause the tests to hang as the open transaction blocks Base.drop_all() in PostgreSQL. Here we have careful instructions to roll back any pending transaction by hand.
    html = render('core/notfound.html', {}, request=request)
    resp = Response(html)
    resp.status_code = 404

    # Hint pyramid_redis_session not to generate any session cookies for this response
    resp.cache_control.public = True

    transaction.abort()
    return resp



"""Default implementation of social login handling."""
# Standard Library
import logging

# Pyramid
from pyramid.httpexceptions import HTTPFound
from pyramid.request import Request
from pyramid.response import Response
from pyramid.settings import aslist
from zope.interface import implementer

from authomatic.adapters import WebObAdapter
from authomatic.core import LoginResult

# Websauna
from websauna.system.core import messages
from websauna.system.user.interfaces import AuthenticationFailure
from websauna.system.user.interfaces import IOAuthLoginService
from websauna.system.user.social import NotSatisfiedWithData
from websauna.system.user.utils import get_authomatic
from websauna.system.user.utils import get_login_service
from websauna.system.user.utils import get_social_login_mapper


logger = logging.getLogger(__name__)


@implementer(IOAuthLoginService)
class DefaultOAuthLoginService:

    def __init__(self, request: Request):
        self.request = request

    def handle_request(self, provider_name) -> Response:
        """Handle all requests coming to login/facebook, login/twitter etc. endpoints.

        * Login form does an empty HTTP POST request to initiate OAuth process

        * Federated authentication service does HTTP GET redirect when they accept OAuth authentication request
        """
        ae = AuthomaticLoginHandler(self.request, provider_name)
        response = ae.handle()
        return response


class AuthomaticLoginHandler:
    """Social login (OAuth/authomatic) internal handling.

   Subclass and override this for customizations.

    Internal implementation of handling OAuth endpoint. The request must be processed as view /login/{provider_name} where provider_name is one of the Authomatic providers set up by Initializer.configure_authomatic(). This view will

    * Check if the request is internal login request and then redirect to OAuth provider

    * Process POST/redirect back from the OAuth Provider

    * Call ISocialAuthMapper to create the user account for incoming social login

    The function returns a tuple. If login success, HTTP response is set. If login fails automatic result is set. If the automatic result is set you are expected to render a login page with error message on it.

    :return: Tuple (HTTP response, Authomatic result)
    """

    def __init__(self, request: Request, provider_name: str):
        """Initialize AuthomaticLoginHandler.

        :param request: Pyramid request.
        :param provider_name: Authomatic provider name.
        """
        self.request = request
        self.provider_name = provider_name

        settings = request.registry.settings
        self.social_logins = aslist(settings.get("websauna.social_logins", ""))

        # Allow only logins which we configured
        assert provider_name in self.social_logins, "Attempt to login non-configured social media {}".format(provider_name)

        self.mapper = get_social_login_mapper(request.registry, provider_name)
        assert self.mapper, "No social media login mapper configured for {}".format(provider_name)

    def process_form(self):
        """Process form values from the internal post request.

        By default this doesn nothing. If your site wants to combine e.g. login + choose product action to single POST you can do it here.

        Example::

            class TreesAuthomaticLoginHandler(AuthomaticLoginHandler):

                def process_form(self):
                    request = self.request
                    product_id = request.POST.get("product_id")
                    if product_id:
                        request.session["delivery_data"] = {
                            "product": product_id,
                            "delivery_details": {},
                            "started": now().isoformat()
                        }
                        request.session.changed()


        """

    def do_success(self, authomatic_result: LoginResult) -> Response:
        """Handle we got a valid OAuth login data.

        Try and log in the user.
        """
        user = self.mapper.capture_social_media_user(self.request, authomatic_result)
        try:
            login_service = get_login_service(self.request)
            return login_service.authenticate_user(user, login_source=self.provider_name)
        except AuthenticationFailure as e:
            messages.add(self.request, kind="error", msg=str(e), msg_id="msg-cannot-login-social-media-user")
            login_url = self.request.route_url("login")
            return HTTPFound(location=login_url)

    def do_error(self, authomatic_result: LoginResult, e: Exception) -> Response:
        """Handle getting error from OAuth provider."""
        # We got some error result, now let see how it goes
        if e:
            # Leave a cue for sysadmins everything is not right. Use INFO level as usually this is just the user pressing Cancel on the OAuth pop up screen.
            logger.info(e)

        messages.add(self.request, kind="error", msg=str(e), msg_id="msg-authomatic-login-error")
        login_url = self.request.route_url("login")
        return HTTPFound(location=login_url)

    def do_bad_request(self):
        """Handle getting HTTP GET to POST endpoint.

        GoogleBot et. al.
        """
        login_url = self.request.route_url("login")
        return HTTPFound(location=login_url)

    def handle(self) -> Response:

        # Login is always state changing operation
        if self.request.method != 'POST' and not self.request.params:
            self.do_bad_request()

        # We will need the response to pass it to the WebObAdapter.
        response = Response()

        # Get handle to Authomatic instance
        authomatic = get_authomatic(self.request.registry)

        # Start the login procedure.
        class InternalPOSTWebObAdapter(WebObAdapter):
            """Having CSRF token in form data messes Authomatic internally so we strip it away with this hack.."""
            @property
            def params(self):
                return dict()

        if "csrf_token" in self.request.POST:
            # This was internal HTTP POST by our own site to this view
            # TODO: Make it use to explicit post parameter to detect this
            self.process_form()
            adapter = InternalPOSTWebObAdapter(self.request, response)
        else:
            adapter = WebObAdapter(self.request, response)

        authomatic_result = authomatic.login(adapter, self.provider_name)

        if not authomatic_result:
            # Guess it wants to redirect us to Facebook et. al and it set response through the adapter
            return response

        # Grab exception if we get any
        e = None

        if not authomatic_result.error:
            # Login succeeded
            try:
                return self.do_success(authomatic_result)
            except NotSatisfiedWithData as ex:
                # TODO: Clean this up - some scoping issues?
                authomatic_result.error = ex
                e = ex

        return self.do_error(authomatic_result, e)



"""Admin URL endpoints for user and group management."""
# Websauna
from websauna.system.admin.modeladmin import ModelAdmin
from websauna.system.admin.modeladmin import model_admin
from websauna.system.crud.urlmapper import Base64UUIDMapper
from websauna.system.user.models import Group
from websauna.system.user.models import User


@model_admin(traverse_id="user")
class UserAdmin(ModelAdmin):
    """Models/user admin root resource."""
    #: Traverse title
    title = "Users"
    singular_name = "user"
    plural_name = "users"
    model = User

    mapper = Base64UUIDMapper()

    class Resource(ModelAdmin.Resource):
        """Wrap one SQLAlchemy user mode to admin resource.

        ``get_object()`` returns :py:class:`websauna.system.user.model.User`.
        """

        def get_title(self) -> str:
            """Return the title to be used on the admin.

            :return: Object friendly_name.
            """
            return self.get_object().friendly_name


@model_admin(traverse_id="group")
class GroupAdmin(ModelAdmin):
    """Models/groups admin root resource."""

    #: Traverse title
    title = "Groups"
    singular_name = "group"
    plural_name = "groups"
    model = Group

    mapper = Base64UUIDMapper()

    class Resource(ModelAdmin.Resource):
        """Wrap one SQLAlchemy group model to admin resource.

        ``get_object()`` returns :py:class:`websauna.system.user.model.Group`.
        """



# Pyramid
from pyramid.events import subscriber

# Websauna
from websauna.system.admin import menu
from websauna.system.admin.events import AdminConstruction


@subscriber(AdminConstruction)
def contribute_admin(event):
    """Add notebook entry to the admin user interface."""
    admin = event.admin
    entry = menu.RouteEntry(
        "admin-notebook",
        label="Shell",
        icon="fa-terminal",
        route_name="admin_shell",
        condition=lambda entry, request: request.has_permission('shell')
    )
    admin.get_quick_menu().add_entry(entry)



"""User friend CSRF error page."""
# Standard Library
import logging

# Pyramid
from pyramid.exceptions import BadCSRFToken
from pyramid.renderers import render
from pyramid.response import Response
from pyramid.view import view_config

# Websauna
from websauna.system.http import Request


logger = logging.getLogger(__name__)


@view_config(context=BadCSRFToken, require_csrf=False)
def bad_csrf_token(context: BadCSRFToken, request: Request):
    """User friendly error page about bad CSRF token."""

    # Log this as a warning
    session = request.session
    token = session.get_csrf_token()
    logger.warn("Bad CSRF error: session: %s IP: %s cookie: %s user agent: %s", request.session.session_id, request.client_addr, token, request.user_agent)

    html = render('core/badcsrftoken.html', {}, request=request)
    resp = Response(html)
    resp.status_code = 400

    # Hint pyramid_redis_session not to generate any session cookies for this response
    resp.cache_control.public = True

    # Make sure nothing is written or no transaction left open on 500
    request.tm.abort()

    return resp



"""Colander schemas."""
# Standard Library
import typing as t

# Pyramid
import colander as c
import deform
import deform.widget as w

# Websauna
from websauna.system.form.schema import CSRFSchema
from websauna.system.form.sqlalchemy import UUIDModelSet
from websauna.system.form.sqlalchemy import convert_query_to_tuples
from websauna.system.user.interfaces import IGroupModel
from websauna.system.user.utils import get_group_class
from websauna.system.user.utils import get_user_class
from websauna.system.user.utils import get_user_registry
from websauna.utils.slug import uuid_to_slug


#: Todo change to setting, convert to deferred widget
PASSWORD_MIN_LENGTH = 6


def defer_group_class(node: c.SchemaNode, kw: dict) -> t.Type[IGroupModel]:
    """Colander helper deferred to assign the current group model.

    :param node: Colander SchemaNode
    :param kw: Keyword arguments.
    :return: IGroupModel
    """
    request = kw.get("request")
    assert request, "To use this widget you must pass request to Colander schema.bind()"

    return get_group_class(request.registry)


def group_vocabulary(node: c.SchemaNode, kw: dict) -> t.List[t.Tuple[str, str]]:
    """Convert all groups on the site to (uuid, name) tuples for checkbox and select widgets.

    :param node: Colander SchemaNode
    :param kw: Keyword arguments.
    :return: List of tuples with uuid, name for all groups on the site.
    """
    Group = defer_group_class(node, kw)
    request = kw["request"]

    def first_column_getter(group: Group):
        return uuid_to_slug(group.uuid)

    return convert_query_to_tuples(request.dbsession.query(Group).all(), first_column=first_column_getter, second_column="name")


def optional_group_vocabulary(node: c.SchemaNode, kw: dict) -> t.List[t.Tuple[str, str]]:
    """Convert all groups on the site to (uuid, name) tuples for checkbox and select widgets. Include a choice for no choice.

    :param node: Colander SchemaNode
    :param kw: Keyword arguments.
    :return: List of tuples with uuid, name for all groups on the site.
    """
    options = group_vocabulary(node, kw)
    return [("", "[ not selected ]")] + options


def validate_unique_user_email(node: c.SchemaNode, value: str, **kwargs: dict):
    """Make sure we cannot enter the same username twice.

    :param node: Colander SchemaNode.
    :param value: Email address.
    :param kwargs: Keyword arguments.
    :raises: c.Invalid if email address already taken.
    """

    request = node.bindings["request"]
    dbsession = request.dbsession
    User = get_user_class(request.registry)
    value = value.strip()
    if dbsession.query(User).filter_by(email=value).one_or_none():
        raise c.Invalid(node, "Email address already taken")


def email_exists(node: c.SchemaNode, value: str):
    """Colander validator that ensures a User exists with the email.

    :param node: Colander SchemaNode.
    :param value: Email address.
    :raises: c.Invalid if email is not registered for an User.
    """
    request = node.bindings['request']
    User = get_user_class(request.registry)
    exists = request.dbsession.query(User).filter(User.email.ilike(value)).one_or_none()
    if not exists:
        raise c.Invalid(node, "Email does not exists: {email}".format(email=value))


class GroupSet(UUIDModelSet):
    """A set of Group objects referred by their uuid."""

    label_column = "name"

    def get_model(self, node: c.SchemaNode) -> t.Type[IGroupModel]:
        """Return Group class.

        :param node: Colander SchemaNode.
        :return: Class implementing IGroupModel.
        """
        request = node.bindings["request"]
        return get_group_class(request.registry)


class RegisterSchema(CSRFSchema):
    """Username-less registration form schema."""

    email = c.SchemaNode(
        c.String(),
        title='Email',
        validator=c.All(c.Email(), validate_unique_user_email),
        widget=w.TextInputWidget(size=40, maxlength=260, type='email'))

    password = c.SchemaNode(
        c.String(),
        validator=c.Length(min=PASSWORD_MIN_LENGTH),
        widget=deform.widget.CheckedPasswordWidget(),
    )


class LoginSchema(CSRFSchema):
    """Login form schema.

    The user can log in both with email and his/her username, though we recommend using emails as users tend to forget their usernames.
    """

    username = c.SchemaNode(c.String(), title='Email', validator=c.All(c.Email()), widget=w.TextInputWidget(size=40, maxlength=260, type='email'))

    password = c.SchemaNode(c.String(), widget=deform.widget.PasswordWidget())


class ResetPasswordSchema(CSRFSchema):
    """Reset password schema."""

    user = c.SchemaNode(
        c.String(),
        missing=c.null,
        widget=deform.widget.TextInputWidget(template='readonly/textinput'))

    password = c.SchemaNode(
        c.String(),
        validator=c.Length(min=2),
        widget=deform.widget.CheckedPasswordWidget()
    )


def validate_user_exists_with_email(node: c.SchemaNode, value: str):
    """Colander validator that ensures a User exists with the email.'

    :param node: Colander SchemaNode.
    :param value: Email address.
    :raises: c.Invalid if email is not registered for an User.
    """
    request = node.bindings['request']

    user_registry = get_user_registry(request)
    user = user_registry.get_by_email(value)

    if not user:
        raise c.Invalid(node, "Cannot reset password for such email: {email}".format(email=value))


class ForgotPasswordSchema(CSRFSchema):
    """Used on forgot password view."""

    email = c.SchemaNode(
        c.Str(),
        title='Email',
        validator=c.All(c.Email(), validate_user_exists_with_email),
        widget=w.TextInputWidget(size=40, maxlength=260, type='email', template="textinput_placeholder"),
        description="The email address under which you have your account. Example: joe@example.com")



"""Permission test views."""

# Pyramid
from pyramid.response import Response

# Websauna
from websauna.system.core.route import simple_route


@simple_route("test_authenticated", permission="authenticated")
def test_authenticated(request):
    return Response("<span id='ok'></span>", content_type="text/html")



"""ws-alembic script.

Wrapper for underlying ``alembic`` script, that is able to handle configuration with includes.
"""
# Standard Library
import sys

# SQLAlchemy
from alembic.util import compat

from pkg_resources import load_entry_point

# Websauna
from websauna.utils.config.includer import IncludeAwareConfigParser


# We need to monkey-patch Alembic ConfigParser
compat.SafeConfigParser = IncludeAwareConfigParser


def main():
    sys.exit(
        load_entry_point('alembic', 'console_scripts', 'alembic')()
    )



"""Facebook login tests.

To run::

    FACEBOOK_USER="x@example.com" FACEBOOK_PASSWORD="y" py.test websauna -s --splinter-webdriver=firefox --splinter-make-screenshot-on-failure=false --ini=test.ini -k test_facebook


"""
# Standard Library
import os

# Pyramid
import transaction

import pytest
from flaky import flaky

# Websauna
from websauna.system.devop.cmdline import init_websauna
from websauna.system.user.models import User
from websauna.system.tests.utils import create_user
from websauna.tests.webserver import customized_web_server


HERE = os.path.dirname(__file__)

#: Selector we use to detect if we are on the Facebook login page
FACEBOOK_LOGIN_TEST_CSS = "#pageFooter"


@pytest.fixture(scope="module")
def facebook_app(request):
    """Construct a WSGI app with tutorial models and admins loaded."""
    ini_file = os.path.join(HERE, "facebook-test.ini")
    request = init_websauna(ini_file)
    return request.app


_web_server = None


@pytest.fixture(scope="module")
def web_server(request, facebook_app):
    """Run a web server with Facebook login settings."""

    # We cache the web server here, so that @flaky doesn't try to recreate it in the same port
    # when Zuckerberg's high quality product fails
    global _web_server

    # customized_port must match one in Facebook app settings
    if not _web_server:
        web_server_factory = customized_web_server(request, facebook_app, customized_port=6662)
        _web_server = web_server_factory()

    return _web_server


def do_facebook_login(browser):
    """Splinter yourself in to the Facebook app."""
    b = browser

    fb_user = os.environ.get("FACEBOOK_USER")
    assert fb_user, "Please configure your Facebook secrets as environment variables to run the tests"
    fb_password = os.environ["FACEBOOK_PASSWORD"]

    assert browser.is_element_present_by_css(FACEBOOK_LOGIN_TEST_CSS)

    # FB login
    b.fill("email", fb_user)
    b.fill("pass", fb_password)
    b.find_by_css("button[name='login']").click()

    # FB allow app confirmation dialog - this is only once per user unless you reset in in your FB profile
    if b.is_text_present("will receive the following info"):
        b.find_by_css("button[name='__CONFIRM__']").click()

    assert b.is_element_present_by_css("#msg-you-are-logged-in")


def do_facebook_login_if_facebook_didnt_log_us_already(browser):
    """Facebook doesn't give us login dialog again as the time is so short, or Authomatic does some caching here?."""

    if browser.is_element_present_by_css(FACEBOOK_LOGIN_TEST_CSS):
        do_facebook_login(browser)
    else:
        # Clicking btn-facebook-login goes directly through to the our login view
        pass


@pytest.mark.skipif(not os.environ.get("FACEBOOK_USER"), reason="Give Facebook user/pass as environment variables")
@flaky
def test_facebook_first_login(web_server, browser, dbsession):
    """Login an user."""

    b = browser
    b.visit(web_server)

    b.click_link_by_text("Sign in")

    assert b.is_element_visible_by_css("#login-form")

    b.find_by_css(".btn-login-facebook").click()
    do_facebook_login_if_facebook_didnt_log_us_already(browser)
    assert b.is_element_present_by_css("#msg-you-are-logged-in")

    # See that we got somewhat sane data
    with transaction.manager:
        assert dbsession.query(User).count() == 1
        u = dbsession.query(User).get(1)
        assert u.first_login
        assert u.email == os.environ["FACEBOOK_USER"]
        assert u.is_admin()  # First user becomes admin
        assert u.activated_at

    b.find_by_css("#nav-logout").click()


@pytest.mark.skipif(not os.environ.get("FACEBOOK_USER"), reason="Give Facebook user/pass as environment variables")
@flaky
def test_facebook_second_login(web_server, browser, dbsession):
    """Login second time through Facebook and see our first_login flag is unset.
    """
    b = browser

    # Initiate Facebook login with Authomatic
    b.visit("{}/login".format(web_server))
    b.find_by_css(".btn-login-facebook").click()

    do_facebook_login_if_facebook_didnt_log_us_already(b)
    assert b.is_text_present("You are now logged in")
    b.find_by_css("#nav-logout").click()

    assert b.is_element_present_by_css("#msg-logged-out")

    # And again!

    b.visit("{}/login".format(web_server))
    b.find_by_css(".btn-login-facebook").click()

    do_facebook_login_if_facebook_didnt_log_us_already(b)

    assert b.is_element_present_by_css("#msg-you-are-logged-in")

    # See that we got somewhat sane data
    with transaction.manager:
        assert dbsession.query(User).count() == 1
        u = dbsession.query(User).get(1)
        assert not u.first_login
        assert u.activated_at

    b.find_by_css("#nav-logout").click()


@pytest.mark.skipif(not os.environ.get("FACEBOOK_USER"), reason="Give Facebook user/pass as environment variables")
@flaky
def test_facebook_login_disabled_user(web_server, browser, dbsession, init):
    """Logged in user which is not enabled should give an error.."""

    with transaction.manager:
        u = create_user(dbsession, init.config.registry, email=os.environ["FACEBOOK_USER"])
        u.enabled = False

    b = browser
    b.visit(web_server)

    b.click_link_by_text("Sign in")

    assert b.is_element_visible_by_css("#login-form")

    b.find_by_css(".btn-login-facebook").click()

    do_facebook_login_if_facebook_didnt_log_us_already(browser)

    assert b.is_element_present_by_css("#msg-cannot-login-social-media-user")



"""
``websauna.system.core.viewconfig`` provides a class decorator ``@view_overrides`` which allows subclasses to partially override parent class view configuration. The primary use case is to have generic class views which you can subclass for more specific use cases. The pattern is very powerful when combined with Pyramid's travesing contexts.

Example:

* There is a generic edit view for all of your models called *GenericEdit*

* But for a specific model, let's say Car, you want to override parts of *GenericEdit* e.g. to include a widget to handle car colour selection. Other models, like House and Person, still use *GenericEdit*.

The code would be::

     from websauna.system.core.viewconfig import view_overrides

     # We define some model structure for which we are going to create edit views
     class BaseModel:
          pass


     class House(BaseModel):
          pass


     class Car(BaseModel):
          pass


     # This is the base edit view class. You can use it as is or subclass it to override parts of it.
     class GenericEdit:

         widgets = ["name", "price"]

         def __init__(self, context, request):
             self.context = context
             self.request = request

         @view_config(name="edit", context=BaseModel)
         def edit(self):
             # Lot's of edit logic code goes here which
             # we don't want to repeat....
             pass


     # This overrides edit() method from GenericEdit.edit() with a different @view_config(context) parameters.
     # Otherwise @view_config() parameters are copied as is.
     @view_overrides(context=Car)
     class CarEdit(GenericEdit):
         widgets = ["name", "price", "color", "year]


      # Some dummy traversing which shows how view are mapped to traversing context
      class Root:
          '''Pyramid's traversing root.'''

          def __init__(self, request):
               pass

          def __getitem__(self, name):
              if is_car(name):
                    return Car(name)
               else:
                    return House(name):


Now one could traverse edit views like::

     /my_house/edit
     /my_car/edit

... and the latter would serve car-specific edit form.

The ``@view_overrides`` pattern can be also used with routing based views to override e.g ``route_name`` and ``renderer`` (the template of subclass view). For those examples, please see testing source code.

The implementation is based on `venusian.lift() <http://venusian.readthedocs.org/en/latest/api.html#venusian.lift>`_ function with the overriding bits added in.
"""
# Standard Library
import sys
from inspect import getmro
from inspect import isclass

# Pyramid
from venusian import ATTACH_ATTR
from venusian import LIFTONLY_ATTR
from venusian import Categories
from venusian.advice import getFrameInfo


def _create_child_view_config_from_parent_cb(cb, module_name, liftid, cscope, overrides):
    """Override view_config declarations from parent for the child.

    Inspect the given venusian decorator callback to see if it is view_config.. If it is view_config, then create equal entry with overridden settings.

    :return: (callback tuple, view_config Ωere overridden flag)
    """

    # Arguments look like:
    # (<function view_config.__call__.<locals>.callback at 0x101d7d268>, 'websauna.system.core.viewconfig.tests.testmodule', 'render None', 'class')

    # Check if we are view_config or previously nested @view_overrides
    if not (cb.__qualname__.startswith("view_config") or cb.__qualname__.startswith("_create_child_view_config_from_parent_cb")):
        return (cb, module_name, liftid, cscope), False

    # OK we are view_config, then pry the decorators parameters out from the closure. If this is not evil Python programming, I don't know what is.
    # Closure looks like (<cell at 0x101cfc948: AttachInfo object at 0x101db4518>, <cell at 0x101cfc888: dict object at 0x101da08c8>)

    if not len(cb.__closure__) == 2:
        # Doesn't look like what we are looking for
        return (cb, module_name, liftid, cscope), False

    # This is like:
    # {'route_name': 'parent', '_info': ('/Users/mikko/code/trees/websauna.system.core.viewconfig/websauna/viewconfig/tests/testmodule.py', 15, 'Parent', '@view_config(route_name="parent", renderer="foobar.html")'), 'renderer': 'foobar.html', 'attr': 'render'}
    info = cb.__closure__[0].cell_contents
    settings = cb.__closure__[1].cell_contents

    # Recreate the callback with new parameters
    settings = settings.copy()
    settings.update(overrides)

    # See pyramid.view.view_config
    def new_callback(context, name, ob):
        config = context.config.with_package(info.module)
        config.add_view(view=ob, **settings)

    return (new_callback, module_name, liftid, cscope), True


class view_overrides(object):
    """A class decorator which overrides chosen view_config arguments from the parent class.

    """
    def __init__(self, categories=None, **kwargs):
        self.categories = categories
        self.kwargs = kwargs

    def __call__(self, wrapped):
        # Shamefully stolen from venusian.lift

        if not isclass(wrapped):
            raise RuntimeError('"view_overrides" only works as a class decorator; you tried to use it against %r' % wrapped)

        frame = sys._getframe(1)
        scope, module, f_locals, f_globals, codeinfo = getFrameInfo(frame)
        module_name = getattr(module, '__name__', None)
        newcategories = Categories(wrapped)
        newcategories.lifted = True

        # Did we find any view_config decorators from the parents of wrapped class
        found = False

        for cls in getmro(wrapped):

            attached_categories = cls.__dict__.get(ATTACH_ATTR, None)
            if attached_categories is None:
                attached_categories = cls.__dict__.get(LIFTONLY_ATTR, None)

            if attached_categories is not None:

                for cname, category in attached_categories.items():
                    if cls is not wrapped:
                        if self.categories and cname not in self.categories:
                            continue
                    callbacks = newcategories.get(cname, [])
                    newcallbacks = []

                    for cb, _, liftid, cscope in category:
                        append = True
                        toappend = (cb, module_name, liftid, cscope)

                        if cscope == 'class':
                            for ncb, _, nliftid, nscope in callbacks:
                                if (nscope == 'class' and liftid == nliftid):
                                    append = False
                        if append:
                            toappend, _found = _create_child_view_config_from_parent_cb(*toappend, overrides=self.kwargs)
                            found = found or _found
                            newcallbacks.append(toappend)
                    newcategory = list(callbacks) + newcallbacks
                    newcategories[cname] = newcategory
                if attached_categories.lifted:
                    break
        if newcategories:  # if it has any keys
            setattr(wrapped, ATTACH_ATTR, newcategories)

        if not found:
            raise RuntimeError('"view_overrides" could not find any @view_config decorators on the parent classes of %r' % wrapped)
        return wrapped



# Pyramid
from pyramid.request import Request


class FlashMessage(object):
    """Internal storage format of flash message."""

    # High traffic sites might have plenty of these objects, so optimize memory usage here a bit
    __slots__ = ('kind', 'plain', 'rich', 'msg_id', 'extra')

    KINDS = set(['error', 'warning', 'info', 'success'])

    def __getstate__(self):
        '''Because we are using __slots__, pickling needs this method.'''
        return {'kind': self.kind, 'plain': self.plain, 'rich': self.rich, 'msg_id': self.msg_id, 'extra': self.extra}

    def __setstate__(self, state):
        self.kind = state.get('kind')
        self.plain = state.get('plain')
        self.rich = state.get('rich')
        self.msg_id = state.get('msg_id')
        self.extra = state.get('extra')

    def __init__(self, plain=None, rich=None, kind='warning', msg_id=None, extra=None):
        """

        :param plain: Message as plain text
        :param rich: Message as HTML
        :param kind: CSS class for the message
        :param msg_id: CSS id and message id
        :param extra: Extra payload (not used by the default renderer)
        """
        assert (plain and not rich) or (rich and not plain)
        assert kind in self.KINDS, "Unknown kind of alert: \"{}\". " \
            "Possible kinds are {}".format(kind, self.KINDS)

        self.kind = kind
        self.rich = rich
        self.plain = plain
        self.msg_id = msg_id
        self.extra = extra

    def __repr__(self):
        return 'FlashMessage("{}")'.format(self.plain)

    def __unicode__(self):
        return self.rich or self.plain

    def get_id(self):
        """Allows session storage to know if this message has been already added."""
        return any((self.msg_id, self.rich, self.plain,))

    # http://stackoverflow.com/a/4901847/315168

    def __hash__(self):
        return hash(self.get_id())

    def __eq__(self, other):
        return self.get_id() == other.get_id()

    def __ne__(self, other):
        # Not strictly necessary, but to avoid having both x==y and x!=y
        # True at the same time
        return not(self == other)


def add(request: Request, msg: str, kind: str="info", msg_id: str=None, extra: dict=None, html=False, allow_duplicates=False):
    """Add a message which is shown to the user on the next page load.

    This is so called Flash message. The message is stored in the session. On the next page load, the HTML templates and framework must read all pending messages from the session store and display them to the user. Usually this is a notification rendered at the top of the page.

    Duplicate messages are discarded based on ``msg_id`` unless ``allow_duplicates`` is set.

    Messages are stored in the session queue, keyed by the message kind.

    :param msg: Message as a string
    :param kind: One of 'error', 'warning', 'info', 'success'. Defaults to 'info'
    :param msg_id: CSS id set on the alert box. Useful for functional testing.
    :param html: The message is HTML
    :param extra: Dictionary of application specific extra parameters or None. The content must be pickable.
    :param allow_duplicates: Allow duplicate messages
    """

    # Don't postpone problems through serialization if the user gives non-str object accidentally
    assert type(msg) == str

    if html:
        msg = FlashMessage(rich=msg, kind=kind, msg_id=msg_id, extra=extra)
    else:
        msg = FlashMessage(plain=msg, kind=kind, msg_id=msg_id, extra=extra)

    request.session.flash(msg, queue=kind, allow_duplicate=allow_duplicates)


def clear(request: Request, kinds=("info", "error", "warning", "success")):
    """Clear all messages from the flash message queue."""

    for kind in kinds:
        request.session.pop_flash(queue=kind)



# Pyramid
from pyramid.response import Response

# Websauna
from websauna.system.core.route import simple_route


# ``simple_route`` decorator can be used to associate view configuration
# information with a function, method, or class that acts as a Pyramid view
# callable.
@simple_route('/path/to/view')
def view_callable(request):
    return Response('Hello')


# You can set route name or any of view configuration arguments explicitly.
@simple_route(
    '/path/to/view/{arg}',
    route_name='foobar',
    request_method='POST',
    renderer='json')
def my_view(request):
    arg = request.matchdict["arg"]
    return {'message': 'Hello {}'.format(arg)}


# More than one decorator can be stacked on top of any number of others. Each
# decorator creates a separate route and view registration. It's better to
# provide ``route_name`` explicitly in this case.
@simple_route('/change', route_name='change')
@simple_route('/edit', route_name='edit')
def edit(request):
    return Response('edited!')


# ``simple_route` can be used as a class decorator.
@simple_route('/class-as-a-view')
class ClassAsAView(object):
    def __init__(self, request):
        self.request = request

    def __call__(self):
        return Response('class-as-a-view')


# The decorator can also be used against a method of a class.
class MyViewClass(object):

    # Class constructor must accept an argument list in one of two forms:
    #   * either a single argument, request, or
    #   * two arguments, context, request.
    def __init__(self, request):
        self.request = request

    # The method which is decorated must return a response.
    @simple_route('/amethod')
    def amethod(self):
        return Response('amethod')


# Using the decorator against a particular method of a class is equivalent to
# using the attr parameter in a decorator attached to the class itself.
@simple_route('/class-attr', attr='amethod')
class MyViewClassAttr(object):

    def __init__(self, request):
        self.request = request

    def amethod(self):
        return Response('class-attr')



"""Permission test views."""
# Standard Library
import typing as t

# Pyramid
from pyramid.interfaces import ILocation
from pyramid.response import Response
from pyramid.view import view_config
from zope.interface import implementer

# Websauna
from websauna.system.core.interfaces import IContainer
from websauna.system.core.root import Root
from websauna.system.core.route import simple_route
from websauna.system.core.sitemap import include_in_sitemap
from websauna.system.core.traversal import Resource
from websauna.system.http import Request


class SampleResource(Resource):
    """Leaf resource in tree."""

    def __init__(self, request: Request, name: str):
        super(SampleResource, self).__init__(request)
        self.name = name


@implementer(IContainer)
class SampleContainer(SampleResource):
    """Node resource in tree."""

    def __init__(self, request, name):
        super(SampleContainer, self).__init__(request, name)

    def items(self) -> t.Iterable[t.Tuple[str, ILocation]]:
        # Every container instan
        request = self.request

        def construct_child(child_id, resource: Resource):
            # Set __parent__ pointer
            resource = Resource.make_lineage(self, resource, child_id)
            return child_id, resource

        # Assume this gets populated dynamically from the database
        # when items() is first time called.
        yield construct_child("foo", SampleResource(request, "Foo"))
        yield construct_child("bar", SampleResource(request, "Bar"))

        # First level container has second level nested container
        if self.name == "Container folder":
            yield construct_child("nested", SampleContainer(request, "Nested"))


@view_config(context=SampleResource, name="", route_name="sitemap_test")
def default_sample_view(sample_resource: SampleResource, request: Request):
    return Response()


@view_config(context=SampleResource, name="additional", route_name="sitemap_test")
def additional_sample_view(sample_resource: SampleResource, request: Request):
    return Response()


def traverse_condition(context, request):
    return True


@view_config(context=SampleResource, name="conditional", route_name="sitemap_test")
@include_in_sitemap(condition=traverse_condition)
def conditional_sample_view(sample_resource: SampleResource, request: Request):
    return Response()


def skipped_condition(context, request):
    return False


@view_config(context=SampleResource, name="skipped_conditional", route_name="sitemap_test")
@include_in_sitemap(condition=skipped_condition)
def skipped_conditional(sample_resource: SampleResource, request: Request):
    return Response()


@view_config(context=SampleContainer, name="", route_name="sitemap_test")
def default_container_view(sample_resource: SampleResource, request: Request):
    return Response()


@view_config(context=SampleContainer, name="additional", route_name="sitemap_test")
def additional_container_view(sample_resource: SampleResource, request: Request):
    return Response()


@view_config(context=SampleContainer, name="permissioned", route_name="sitemap_test", permission="edit")
def permissioned_container_view(sample_resource: SampleResource, request: Request):
    return Response()


@simple_route("/parameter_free_route", route_name="parameter_free_route")
def parameter_free_route(request: Request):
    return Response()


@simple_route("/parameterized_route/{param}", route_name="parameterized_route")
def parameterized_route(request: Request):
    return Response()


@simple_route("/permissioned_route", route_name="permissioned_route", permission="edit")
def permissioned_route(request: Request):
    return Response()


@simple_route("/post_only_route", route_name="post_only_route", request_method="POST")
def post_only_route(request: Request):
    return Response()


@simple_route("/included_route", route_name="included_route")
@include_in_sitemap(True)
def included_route(request: Request):
    return Response()


@simple_route("/skipped_route", route_name="skipped_route")
@include_in_sitemap(False)
def skipped_route(request: Request):
    return Response()


def condition(context, request):
    return True


@simple_route("/conditional_route", route_name="conditional_route")
@include_in_sitemap(condition=condition)
def another_skipped_route(request: Request):
    return Response()


def container_factory(request):
    """Set up __parent__ and __name__ pointers required for traversal for container root."""
    container = SampleContainer(request, "Container folder")
    root = Root.root_factory(request)
    return Resource.make_lineage(root, container, "container")



"""Handle incoming user events."""
# Pyramid
from pyramid.events import subscriber

# Websauna
from websauna.system.user.events import UserAuthSensitiveOperation
from websauna.utils.time import now


@subscriber(UserAuthSensitiveOperation)
def user_auth_details_changes(event: UserAuthSensitiveOperation):
    """Default logic how to invalidate sessions on user auth detail changes.

    If you are using different session management model you can install a custom handle.

    :param event: Incoming event instance
    """
    user = event.user
    # Update the timestamp which session validation checks on every request
    user.last_auth_sensitive_operation_at = now()



"""ws-create-user script.

Create a new site user from command line.
"""
# Standard Library
import getpass
import os
import sys
import typing as t

# Websauna
from websauna.system.devop.cmdline import init_websauna
from websauna.system.devop.scripts import feedback_and_exit
from websauna.system.devop.scripts import get_config_uri
from websauna.system.user.events import UserCreated
from websauna.system.user.interfaces import IUserModel
from websauna.system.user.models import Group
from websauna.system.user.utils import get_user_class
from websauna.system.user.utils import get_user_registry
from websauna.utils.time import now


def usage_message(argv: t.List[str]):
    """Display usage message and exit.

    :param argv: Command line arguments.
    :raises sys.SystemExit:
    """
    cmd = os.path.basename(argv[0])
    msg = (
        'usage: {cmd} <config_uri> [password] [--admin]\n'
        '(example: "{cmd} ws://conf/production.ini mikko@example.com verysecret --admin")'
    ).format(cmd=cmd)
    feedback_and_exit(msg, status_code=1, display_border=False)


def create(
        request,
        username: str,
        email: str,
        password: t.Optional[str]=None,
        source: str='command_line',
        admin: bool=False
) -> IUserModel:
    """Create a new site user from command line.

    :param request:
    :param username: Username, usually an email.
    :param email: User's email.
    :param password: Password.
    :param source: Source of this user, in here, command_line.
    :param admin: Set this user to admin. The first user is always implicitly admin.
    :return: Newly created user.
    """
    User = get_user_class(request.registry)
    dbsession = request.dbsession
    u = dbsession.query(User).filter_by(email=email).first()
    if u is not None:
        return u

    u = User(email=email, username=username)
    dbsession.add(u)
    dbsession.flush()  # Make sure u.user_data is set

    if password:
        user_registry = get_user_registry(request)
        user_registry.set_password(u, password)

    u.registration_source = source
    u.activated_at = now()

    request.registry.notify(UserCreated(request, u))
    if admin:
        group = dbsession.query(Group).filter_by(name='admin').one_or_none()
        group.users.append(u)

    return u


def main(argv: t.List[str]=sys.argv):
    """Create a new site user from command line.

    :param argv: Command line arguments, second one needs to be the uri to a configuration file.
    :raises sys.SystemExit:
    """
    if len(argv) < 3:
        usage_message(argv)

    config_uri = get_config_uri(argv)

    request = init_websauna(config_uri)
    email = argv[2]
    is_admin = True if '--admin' in argv else False
    password = argv[3] if len(argv) >= 4 and argv[3] != '--admin' else ''

    if not password:
        password = getpass.getpass('Password:')
        password2 = getpass.getpass('Password (again):')
        if password != password2:
            feedback_and_exit('Passwords did not match', display_border=False)

    with request.tm:
        u = create(request, email=email, username=email, password=password, admin=is_admin)
        message = 'Created user #{id}: {email}, admin: {is_admin}'.format(
            id=u.id,
            email=u.email,
            is_admin=u.is_admin()
        )

    feedback_and_exit(message, status_code=None, display_border=True)


if __name__ == "__main__":
    main()



"""Map URL traversing ids to database ids and vice versa."""
# Standard Library
import abc

# Websauna
from websauna.utils import slug
from websauna.utils.slug import SlugDecodeError


class CannotMapException(Exception):
    """We could not extract id from an object."""


class Mapper(abc.ABC):
    """Define mapping interface used by CRUD subsystem."""

    @abc.abstractmethod
    def get_path_from_object(self, obj):
        """Map database object to an travesable URL path."""
        raise NotImplementedError()

    @abc.abstractmethod
    def get_id_from_path(self, path):
        """Map traversable resource name to an database object id."""
        raise NotImplementedError()


class IdMapper(Mapper):
    """Use object/column attribute id to map functions.

    By default this is set to use integer ids, but you can override properties.
    """

    #: What is the object attibute name defining the its id
    mapping_attribute = "id"

    #: Function to translate obj attribute -> URL path str
    transform_to_path = str

    #: Function to translate URL path str -> object id
    transform_to_id = int

    #: is_id(path) function checks whether the given URL path should be mapped to object and is a valid object id. Alternatively, if the path doesn't look like an object id, it could be a view name. Because Pyramid traversing checks objects prior views, we need to let bad object ids to fall through through KeyError, so that view matching mechanism kicks in. By default we check for number integer id.
    #: Some object paths cannot be reliable disquished from view names, like UUID strings. In this case ``is_id`` is None, the lookup always first goes to the database. The database item is not with view name a KeyError is triggerred and thus Pyramid continues to view name resolution. This behavior is suboptimal and may change in the future versions
    is_id = staticmethod(lambda value: value.isdigit())

    def __init__(self, mapping_attribute=None, transform_to_path=None, transform_to_id=None, is_id=None):
        if mapping_attribute:
            self.mapping_attribute = mapping_attribute

        if transform_to_path:
            self.transform_to_path = transform_to_path

        if transform_to_id:
            self.transform_to_id = transform_to_id

        if is_id:
            self.is_id = is_id

    def get_path_from_object(self, obj):

        if not hasattr(obj, self.mapping_attribute):
            raise CannotMapException(
                "Could not find attribute {attr} on object {obj}. The default behavior is to look for attribute/column uuid. If you need to change this behavior define mapper in your CRUD class.".format(
                    attr=self.mapping_attribute,
                    obj=obj
                )
            )

        return self.transform_to_path(getattr(obj, self.mapping_attribute))

    def get_id_from_path(self, path):
        return self.transform_to_id(path)


class Base64UUIDMapper(IdMapper):
    """Map objects to URLs using their UUID property."""

    #: Override this if you want to change the column name containing uuid
    mapping_attribute = "uuid"

    #: Use utils.slug package to produce B64 strings from UUIDs
    transform_to_id = staticmethod(slug.slug_to_uuid)

    #: Use utils.slug package to produce B64 strings from UUIDs
    transform_to_path = staticmethod(slug.uuid_to_slug)

    @staticmethod
    def is_id(val):
        """Try guess if the value is valid base64 UUID slug or not.

        Note that some view names can be valid UUID slugs, thus we might hit database in any case for the view lookup.
        """
        try:
            slug.slug_to_uuid(val)
            return True
        except SlugDecodeError:
            # bytes is not 16-char string
            return False



"""Describe different form edit modes."""

# Standard Library
from enum import Enum


class EditMode(Enum):
    """Different edit modes for where a form can be."""

    #: Generate form for viewing contents (read-only)
    show = 0

    #: Generated form for creating a new object
    add = 1

    #: Generated form for editing an existing object
    edit = 2



"""Paginator."""
# Standard Library
import math
import typing as t
from urllib.parse import parse_qsl
from urllib.parse import urlencode
from urllib.parse import urlsplit
from urllib.parse import urlunsplit


def merge_url_qs(url: str, **kw) -> str:
    """Merge the query string elements of a URL with the ones in ``kw``.

    If any query string element exists in ``url`` that also exists in ``kw``, replace it.

    :param url: An URL.
    :param kw: Dictionary with keyword arguments.
    :return: An URL with keyword arguments merged into the query string.
    """
    segments = urlsplit(url)
    extra_qs = [
        (k, v)
        for (k, v) in parse_qsl(segments.query, keep_blank_values=1)
        if k not in kw
    ]
    qs = urlencode(sorted(kw.items()))
    if extra_qs:
        qs += '&' + urlencode(extra_qs)
    return urlunsplit((segments.scheme, segments.netloc, segments.path, qs, segments.fragment))


class Batch:
    """Present one paginator batch in the list rendering output.

    Originally courtesy of SubstanceD project.

    Given a sequence named ``seq``, and a Pyramid request, return an
    object with the following attributes:

    ``items``

      A list representing a slice of ``seq``.  It will contain the number of
      elements in ``request.params['batch_size']`` or the ``default_size``
      number if such a key does not exist in request.params or the key is
      invalid.  The slice will begin at ``request.params['batch_num']`` or
      zero if such a key does not exist in ``request.params`` or the
      ``batch_num`` key could not successfully be converted to a positive
      integer.

      This value can be iterated over via the ``__iter__`` of the batch
      object.

    ``size``

      The value obtained from ``request.params['batch_size']`` or
      ``default_size`` if no ``batch_size`` parameter exists in
      ``request.params`` or the ``batch_size`` parameter could not
      successfully be converted to a positive interger.

    ``num``

      The value obtained from ``request.params['batch_num']`` or ``0`` if no
      ``batch_num`` parameter exists in ``request.params`` or if the
      ``batch_num`` parameter could not successfully be converted to a
      positive integer.  Batch numbers are indexed from zero, so batch ``0``
      is the first batch, batch ``1`` the second, and so forth.

    ``length``

      This is length of the current batch.  It is usually equal to ``size``
      but may be different in the very last batch.  For example, if the
      ``seq`` is ``[1,2,3,4]`` and the batch size is ``3``, the first batch's
      ``length`` will be ``3`` because the batch content will be ``[1,2,3]``;
      but the second and final batch's ``length`` will be ``1`` because the
      batch content will be ``[4]``.

    ``last``

      The batch number computed from the sequence length of the last batch
      (indexed from zero).

    ``first_url``

      The URL of the first batch.  This will be a URL with ``batch_num`` and
      ``batch_size`` in its query string.  The base URL will be taken from
      the ``url`` value passed to this function.  If a ``url`` value is not
      passed to this function, the URL will be taken from ``request.url``.
      This value will be ``None`` if the current ``batch_num`` is 0.

    ``prev_url``

      The URL of the previous batch.  This will be a URL with ``batch_num``
      and ``batch_size`` in its query string.  The base URL will be taken
      from the ``url`` value passed to this function.  If a ``url`` value is
      not passed to this function, the URL will be taken from
      ``request.url``.  This value will be ``None`` if there is no previous
      batch.

    ``next_url``

      The URL of the next batch.  This will be a URL with ``batch_num`` and
      ``batch_size`` in its query string.  The base URL will be taken from
      the ``url`` value passed to this function.  If a ``url`` value is not
      passed to this function, the URL will be taken from ``request.url``.
      This value will be ``None`` if there is no next batch.

    ``last_url``

      The URL of the next batch.  This will be a URL with ``batch_num`` and
      ``batch_size`` in its query string.  The base URL will be taken from
      the ``url`` value passed to this function.  If a ``url`` value is not
      passed to this function, the URL will be taken from ``request.url``.
      This value will be ``None`` if there is no next batch.

    ``required``

      ``True`` if either ``next_url`` or ``prev_url`` are ``True`` (meaning
      batching is required).

    ``multicolumn``

      ``True`` if the current view should be rendered in multiple columns.

    ``toggle_url``

      The URL to be used for the multicolumn/single column toggle button. The
      ``batch_size``, ``batch_num``, and ``multicolumn`` parameters are
      converted to their multicolumn or single column equivalents. If a user
      is viewing items 40-80 in multiple columns, the toggle will switch to
      items 40-50 in a single column. If a user is viewing items 50-60 in a
      single column, the toggle will switch to items 40-80 in multiple columns.

    ``toggle_text``

      The text to display on the multi-column/single column toggle.

    ``make_columns``

      A method to split ``items`` into a nested list representing columns.

    ``seqlen``

      This is total length of the sequence (across all batches).

    ``startitem``

      The item number that starts this batch (indexed from zero).

    ``enditem``

      The item number that ends this batch (indexed from zero).

    """
    def __init__(self, seq, request, url=None, default_size=10, toggle_size=40,
                 seqlen=None):
        if url is None:
            url = request.url

        try:
            num = int(request.params.get('batch_num', 0))
        except (TypeError, ValueError):
            num = 0
        if num < 0:
            num = 0

        try:
            size = int(request.params.get('batch_size', default_size))
        except (TypeError, ValueError):
            size = default_size
        if size < 1:
            size = default_size

        multicolumn = request.params.get('multicolumn', '') == 'True'

        # create multicolumn/single column toggle attributes
        if multicolumn:
            toggle_num = size * num / default_size
            toggle_size = default_size
            toggle_text = 'Single column'
        else:
            toggle_num = size * num / toggle_size
            toggle_text = 'Multi-column'

        if seqlen is None:
            # won't work if seq is a generator
            seqlen = len(seq)
        start = num * size
        end = start + size
        if end > seqlen:
            end = seqlen

        # normal list slicing is mucho faster than islice
        items = seq[start:end]

        length = len(items)
        last = int(math.ceil(seqlen / float(size)) - 1)

        first_url = None
        prev_url = None
        next_url = None
        last_url = None
        toggle_url = None

        if num:
            first_url = merge_url_qs(url, batch_size=size, batch_num=0)
        if start >= size:
            prev_url = merge_url_qs(url, batch_size=size, batch_num=num - 1)
        if seqlen > end:
            next_url = merge_url_qs(url, batch_size=size, batch_num=num + 1)
        if size and (num < last):
            last_url = merge_url_qs(url, batch_size=size, batch_num=last)

        if prev_url or next_url:
            toggle_url = merge_url_qs(
                url,
                batch_size=toggle_size,
                batch_num=toggle_num,
                multicolumn=not multicolumn,
            )

        self.startitem = start
        self.enditem = end - 1
        self.last = last
        self.seqlen = seqlen
        self.items = items
        self.num = num
        self.size = size
        self.length = length
        self.required = bool(prev_url or next_url)
        self.multicolumn = multicolumn
        self.toggle_url = toggle_url
        self.toggle_text = toggle_text
        self.first_url = first_url
        self.prev_url = prev_url
        self.next_url = next_url
        self.last_url = last_url

    def make_columns(self, column_size=10, num_columns=4):
        """ Break ``self.items`` into a nested list representing columns."""
        columns = []
        for i in range(num_columns):
            start = i * column_size
            end = start + column_size
            part = self.items[start:end]
            columns.append(part)
        return columns

    def __iter__(self):
        return iter(self.items)

    def __len__(self):
        return self.length

    def __nonzero__(self):
        return True

    __bool__ = __nonzero__  # py3


class DefaultPaginator:
    """Default pagination implementation for CRUD, having 20 items per page."""

    template = 'crud/paginator.html'

    default_size = 20

    def __init__(self, template: t.Optional[str]=None, default_size: t.Optional[int]=None):
        """Initialize DefaultPaginator.

        :param template: Path to paginator template.
        :param default_size: Pagination size.
        """
        if template:
            self.template = template

        if default_size:
            self.default_size = default_size

    def paginate(self, seq, request, count, url=None) -> Batch:
        batch = Batch(seq, request, seqlen=count, url=url, default_size=self.default_size)
        return batch



"""Various Colander schema base classes and utililties for forms."""
# Standard Library
import enum
import json
import typing as t

# Pyramid
import colander
import deform  # noQA

#: Backwards compatibility
from .csrf import CSRFSchema  # noQA
from .csrf import add_csrf  # noQA


def validate_json(node, value, **kwargs):
    """Make sure the string content is valid JSON."""

    try:
        json.loads(value)
    except json.JSONDecodeError:
        raise colander.Invalid(node, "Not valid JSON")


def enum_values(source: enum.Enum, default: t.Optional[t.Tuple]=("", "Please choose"), name_transform=str.title) -> t.Iterable[t.Tuple]:
    """Turn Python Enum to key-value pairs lists to be used with selection widgets."""

    def inner():
        if default:
            yield default

        for name, member in source.__members__.items():
            yield (name, name_transform(name))

    return list(inner())


def dictify(schema: colander.Schema, obj: object, excludes=()) -> dict:
    """ Return a dictified version of `obj` using schema information.

    The schema will be used to choose what attributes will be
    included in the returned dict.

    Thus, the return value of this function is suitable for consumption
    as a ``Deform`` ``appstruct`` and can be used to pre-populate
    forms in this specific use case.

    Arguments/Keywords

    :param obj:
        An object instance to be converted to a ``dict`` structure.
        This object should conform to the given schema.  For
        example, ``obj`` should be an instance of this schema's
        mapped class, an instance of a sub-class, or something that
        has the same attributes.

    :param exludes: List of node names not to automatically dictify

    Based on colanderalchemy implemetation.
    """
    dict_ = {}

    for node in schema.children:

        if not getattr(node, "dictify_by_default", True):
            # This is set by CSRF field, so that we don't need to manually
            # skip it between object serializations.
            # Not elegant, but convenient.
            continue

        name = node.name
        if name in excludes:
            continue

        value = getattr(obj, name)

        # SQLAlchemy mostly converts values into Python types
        #  appropriate for appstructs, but not always.  The biggest
        #  problems are around `None` values so we're dealing with
        #  those here.  All types should accept `colander.null` so
        #  we mostly change `None` into that.

        if value is None:
            if isinstance(node.typ, colander.String):
                # colander has an issue with `None` on a String type
                #  where it translates it into "None".  Let's check
                #  for that specific case and turn it into a
                #  `colander.null`.
                dict_[name] = colander.null
            else:
                # A specific case this helps is with Integer where
                #  `None` is an invalid value.  We call serialize()
                #  to test if we have a value that will work later
                #  for serialization and then allow it if it doesn't
                #  raise an exception.  Hopefully this also catches
                #  issues with user defined types and future issues.
                try:
                    node.serialize(value)
                except Exception:
                    dict_[name] = colander.null
                else:
                    dict_[name] = value
        else:
            dict_[name] = value

    return dict_


def objectify(schema: colander.Schema, appstruct: dict, context: object, excludes=()):
    """Store incoming appstruct data on an object like SQLAlchemy model instance"""

    for node in schema.children:

        if not getattr(node, "dictify_by_default", True):
            # This is set by CSRF field, so that we don't need to manually
            # skip it between object serializations.
            # Not elegant, but convenient.
            continue

        name = node.name
        if name in excludes:
            continue

        try:
            setattr(context, name, appstruct[name])
        except AttributeError as e:
            raise AttributeError("Could not set attr {} on {} from schema {}".format(name, context, schema)) from e



"""Command line scripts."""
# Standard Library
import os
import sys
import typing as t
from textwrap import dedent

from pkg_resources import load_entry_point

# Websauna
from websauna.system import SanityCheckFailed
from websauna.system.devop.cmdline import prepare_config_uri


FAIL_MSG = """Sanity check failed

Reason:
{exception}
"""

DEPRECATION_MSG = """
DEPRECATION WARNING

This script is deprecated and will be removed in a future release of Websauna.
Instead, please use the following command;
{script} {config_uri}
"""


def feedback(message: str, display_border: bool=True):
    """Display a feedback message on the console then exit.

    :param message: Message to be displayed to the user.
    :raises sys.SystemExit:
    """
    if display_border:
        border = '=' * 100
        message = '{border}\n{message}\n{border}'.format(border=border, message=message)
    message = dedent(message)
    print(message)


def feedback_and_exit(message: str, status_code: t.Optional[int]=None, display_border: bool=True):
    """Display a feedback message on the console then exit.

    :param message: Message to be displayed to the user.
    :param status_code: Status code to be raised after displaying the message.
    :raises sys.SystemExit:
    """
    feedback(message, display_border)
    sys.exit(status_code)


def get_config_uri(argv: t.List[str]) -> str:
    """Return the config_uri from command line argv.

    :param argv: Sequence of command line string arguments.
    :return: config_uri, i.e: ws://websauna/conf/test.ini
    """
    config_uri = argv[1]
    return prepare_config_uri(config_uri)


def usage_message(argv: t.List[str], additional_params: str='', additional_line: t.Optional[str]=None):
    """Display usage message and exit.

    :param argv: Command line arguments.
    :param additional_params: Additional parameters to be displayed. i.e.: [var=value].
    :param additional_line: Additional line to be added to the end of the message. i.e.: [var=value].
    :raises sys.SystemExit:
    """
    cmd = os.path.basename(argv[0])
    msg = 'usage: {cmd} <config_uri> {params}\n(example: "{cmd} ws://conf/production.ini{line}")'.format(
        cmd=cmd,
        params=additional_params,
        line='' if not additional_line else '\n{0}'.format(additional_line)
    )
    feedback_and_exit(msg, status_code=1, display_border=False)


def display_deprecation_warning(script: str, config_uri: str):
    """Display a deprecation message.

    :param script: Name of the original Pyramid script.
    :param config_uri: URI for the configuration file.
    """
    message = DEPRECATION_MSG.format(script=script, config_uri=config_uri)
    feedback(message, True)


def proxy_to_pyramid_script(script: str, argv: t.List[str]):
    """Proxy call to the original Pyramid script.

    :param script: Name of the original Pyramid script.
    :param argv: Command line arguments.
    :raises sys.SystemExit:
    """
    if len(argv) < 2:
        usage_message(argv)

    # Make sure we are prefixing calls with our plaster schema loader.
    config_uri = get_config_uri(argv)
    argv[1] = config_uri
    display_deprecation_warning(script, config_uri)
    try:
        sys.exit(
            load_entry_point('pyramid', 'console_scripts', script)()
        )
    except SanityCheckFailed as exc:
        feedback_and_exit(FAIL_MSG.format(exception=str(exc)), 1)



"""Test view permission as functional tests."""
# Pyramid
import transaction

# SQLAlchemy
from sqlalchemy.orm import Session

import pytest
from webtest import TestApp   # noQA

# Websauna
import websauna
from websauna.system.tests.utils import create_user
from websauna.system.tests.utils import login
from websauna.tests.webserver import customized_web_server

from . import permissionsamples


@pytest.fixture(scope="module")
def permission_app(request, paster_config):
    '''Custom WSGI app with permission test views enabled.'''

    class Initializer(websauna.system.DemoInitializer):

        def configure_views(self):
            super(Initializer, self).configure_views()
            self.config.scan(permissionsamples)

        def configure_csrf(self):
            """Disable CSRF for this test run for making testing simpler."""
            self.config.set_default_csrf_options(require_csrf=False)

    global_config, app_settings = paster_config
    init = Initializer(global_config, app_settings)
    init.run()
    app = init.make_wsgi_app()
    app.init = init
    return app


@pytest.fixture(scope="module")
def web_server(request, permission_app):
    """Run a web server with custom permission test views enabled."""

    web_server_factory = customized_web_server(request, permission_app)
    _web_server = web_server_factory()
    return _web_server


def test_autheticated_anonymous_not_allowed(web_server, dbsession: Session, init, browser):
    """Anonymous access is not allowed to the views protected by authenticate permission."""
    b = browser
    b.visit("{}/test_authenticated".format(web_server))
    assert b.is_element_present_by_css("#forbidden")


def test_logged_in_has_authenticated_permission(web_server, dbsession: Session, browser, permission_app):
    """Logged in users can access views behind authenticated permission."""

    b = browser

    with transaction.manager:
        create_user(dbsession, permission_app.init.config.registry)

    b.visit(web_server)
    login(web_server, b)

    # Logged in user can access
    b.visit("{}/test_authenticated".format(web_server))
    assert b.is_element_present_by_css("#ok")



"""A broken module where we try to override view class which does not defined views."""
# Websauna
from websauna.system.core.viewconfig import view_overrides


class ParentResource:
    id = "parent"


class ChildResource(ParentResource):
    id = "child"


class ParentView:

    def __init__(self, context, request):
        self.context = context
        self.request = request

    # No views defined


@view_overrides(context=ChildResource)
class ChildView(ParentView):
    pass


class Root:
    """Traversing root."""

    def __init__(self, request):
        pass

    def __getitem__(self, name):
        if name == "parent":
            return ParentResource()
        elif name == "child":
            return ChildResource()



"""Test SQL transaction conflict resolution."""
# Standard Library
import threading
import time

# Pyramid
import transaction

# SQLAlchemy
from sqlalchemy import Column
from sqlalchemy import Integer
from sqlalchemy import Numeric

import pytest

# Websauna
from websauna.system.model.meta import Base
from websauna.system.model.meta import create_dbsession
from websauna.system.model.retry import CannotRetryAnymore
from websauna.system.model.retry import is_retryable
from websauna.system.model.retry import retryable


_test_model = None


def get_test_model():
    global _test_model
    if _test_model:
        return _test_model

    class TestModel(Base):
        """A sample SQLAlchemy model to demostrate db conflicts. """

        __tablename__ = "test_model"

        #: Running counter used in foreign key references
        id = Column(Integer, primary_key=True)

        #: The total balance of this wallet in the minimum unit of cryptocurrency
        #: NOTE: accuracy checked for Bitcoin only
        balance = Column(Numeric(21, 8))

    _test_model = TestModel
    return _test_model


class ConflictThread(threading.Thread):
    """Launch two of these and they should cause database conflict."""

    def __init__(self, session_factory):
        self.session_factory = session_factory
        self.failure = None
        threading.Thread.__init__(self)

    def run(self):

        txn = None

        try:

            dbsession = self.session_factory()

            txn = dbsession.transaction_manager.begin()
            TestModel = get_test_model()

            # Both threads modify the same wallet simultaneously
            w = dbsession.query(TestModel).get(1)
            w.balance += 1

            # Let the other session to start its own transaction
            time.sleep(1)
            dbsession.transaction_manager.commit()
        except Exception as e:
            self.failure = (txn, dbsession, e)


class ConflictResolverThread(threading.Thread):
    """Launch two of these and they should cause database conflict and then retryable resolves it."""

    def __init__(self, dbsession_factory):
        self.dbsession_factory = dbsession_factory
        self.failure = None
        threading.Thread.__init__(self)
        self.success_count = 0
        self.failure_count = 0
        self.retry_count = 0

    def run(self):

        dbsession = self.dbsession_factory()

        # Execute the conflict sensitive code inside a managed transaction
        @retryable(tm=dbsession.transaction_manager)
        def myfunc():

            TestModel = get_test_model()

            # Both threads modify the same wallet simultaneously
            w = dbsession.query(TestModel).get(1)
            w.balance += 1

            # Let the other session to start its own transaction
            time.sleep(1)

        try:
            myfunc()
            self.success_count += 1
        except Exception as e:
            self.failure = e
            self.failure_count += 1

        # See retryable()
        self.retry_count = dbsession.transaction_manager.latest_retry_count


@pytest.fixture
def test_instance(dbsession):

    TestModel = get_test_model()
    Base.metadata.create_all(tables=[TestModel.__table__], bind=dbsession.get_bind())

    with dbsession.transaction_manager:

        # Create an wallet with balance of 10
        w = dbsession.query(TestModel).get(1)
        if not w:
            w = TestModel()
            dbsession.add(w)

        w.balance = 10


@pytest.fixture
def dbsession_factory(test_request):

    if test_request.dbsession.bind.dialect.name == "sqlite":
        # These tests work only for PSQL
        pytest.skip('These can be run only on PostgreSQL')

    def factory():
        dbsession = create_dbsession(test_request.registry, manager=None)
        # Retry each transaction max 1 times
        dbsession.transaction_manager.retry_attempt_count = 2
        return dbsession

    return factory


def test_sql_transaction_conflict(test_instance, dbsession_factory):
    """Run database to a transaction conflict and see what exception it spits out, and make sure we know this is the exception we expect."""

    t1 = ConflictThread(dbsession_factory)
    t2 = ConflictThread(dbsession_factory)

    t1.start()
    t2.start()

    t1.join()
    t2.join()

    # Either thread spits out:
    # sqlalchemy.exc.OperationalError: (TransactionRollbackError) could not serialize access due to concurrent update
    #  'UPDATE btc_wallet SET updated_at=%(updated_at)s, balance=%(balance)s WHERE btc_wallet.id = %(btc_wallet_id)s' {'btc_wallet_id': 1, 'balance': Decimal('11.00000000'), 'updated_at': datetime.datetime(2014, 12, 18, 1, 53, 58, 583219)}
    failure = t1.failure or t2.failure or None
    assert failure is not None

    txn, dbsession, exc = failure
    error = exc

    assert txn.status == "Commit failed"
    assert is_retryable(txn, error) is True


def test_conflict_resolved(test_instance, dbsession_factory, dbsession):
    """Use conflict resolver to resolve conflict between two transactions and see code retry is correctly run."""

    TestModel = get_test_model()

    t1 = ConflictResolverThread(dbsession_factory)
    t2 = ConflictResolverThread(dbsession_factory)

    t1.start()
    t2.start()

    t1.join()
    t2.join()

    # sqlalchemy.exc.OperationalError: (TransactionRollbackError) could not serialize access due to concurrent update
    #  'UPDATE btc_wallet SET updated_at=%(updated_at)s, balance=%(balance)s WHERE btc_wallet.id = %(btc_wallet_id)s' {'btc_wallet_id': 1, 'balance': Decimal('11.00000000'), 'updated_at': datetime.datetime(2014, 12, 18, 1, 53, 58, 583219)}
    failure = t1.failure or t2.failure or None
    assert failure is None  # Both threads pass

    # Check both increments came through
    with transaction.manager:
        w = dbsession.query(TestModel).get(1)
        assert w.balance == 12

    success = sum([t1.success_count, t2.success_count])
    retries = sum([t1.retry_count, t2.retry_count])
    errors = sum([t1.failure_count, t2.failure_count])

    assert success == 2
    assert retries == 1  # At least one thread needs to retry
    assert errors == 0


def test_conflict_some_other_exception(dbsession):
    """See that unknown exceptions are correctly reraised by managed_transaction."""

    @retryable(tm=dbsession.transaction_manager)
    def do_stuff():
        raise ValueError("Unknown exception")

    with pytest.raises(ValueError):
        do_stuff()


def test_give_up(test_instance, dbsession_factory, dbsession):
    """See that the conflict resolver gives up after using given number of attempts to replay transactions."""

    # The resolved has retry count of 1,
    # First t1 success, t2 and t3 clases
    # Then t2 success, t3 retries but is out of
    t1 = ConflictResolverThread(dbsession_factory)
    t2 = ConflictResolverThread(dbsession_factory)
    t3 = ConflictResolverThread(dbsession_factory)

    t1.start()
    t2.start()
    t3.start()

    t1.join()
    t2.join()
    t3.join()

    failure = t1.failure or t2.failure or t3.failure or None
    assert isinstance(failure, CannotRetryAnymore)

    success = sum([t1.success_count, t2.success_count, t3.success_count])
    retries = sum([t1.retry_count, t2.retry_count, t3.retry_count])
    errors = sum([t1.failure_count, t2.failure_count, t3.failure_count])

    assert success == 2
    assert retries == 2
    assert errors == 1

    # Check 2 increments came through
    TestModel = get_test_model()
    with transaction.manager:
        w = dbsession.query(TestModel).get(1)
        assert w.balance == 12



"""Route related helpers."""
# Standard Library
import typing as t

# Pyramid
import venusian
from pyramid.config import Configurator

# Websauna
from websauna.system.http import Request

from .simpleroute import add_simple_route


class simple_route(object):
    """A set of simple defaults for Pyramid routing.

    Pyramid's URL dispatch has separate concepts for routes and views. This gives additional flexibility in that you can one route map to multiple views, using different predicates (e.g.: predicates depending on Accept header, whether request is XHR or not, etc.). In many applications, this flexibility is not needed and having both routes and views adds a bit of complexity and duplication, and reduces DRYness. This module implements some easy-to-use mechanisms that create a route and a view in one step, resulting in simpler, easier to understand code. This kind of makes Pyramid's routing look a bit more like Flask, albeit without Flask's controversial thread locals.

    .. note:: At the moment ``simple_route`` doesn't support ``viewdefaults``.

    Examples:

    .. literalinclude:: ../../../websauna/tests/viewconfig/simple_route_tests_views.py

    Respectively URLs for registered views are:

    .. literalinclude:: ../../../websauna/tests/test_simple_route.py
        :pyobject: test_registered_routes_docs
        :dedent: 11
        :lines: 4-

    """
    def __init__(self, path, *args, **kwargs):
        """Constructor just here to accept parameters for decorator"""
        self.path = path
        self.args = args
        self.kwargs = kwargs

    def __call__(self, wrapped):
        """Attach the decorator with Venusian"""
        args = self.args
        kwargs = self.kwargs

        def callback(scanner, _name, wrapped):
            """Register a view; called on config.scan"""
            config = scanner.config

            # Default to not appending slash
            if "append_slash" not in kwargs:
                kwargs['append_slash'] = False

            # pylint: disable=W0142
            add_simple_route(config, self.path, wrapped, *args, **kwargs)

        info = venusian.attach(wrapped, callback)

        if info.scope == 'class':  # pylint:disable=E1101
            # if the decorator was attached to a method in a class, or
            # otherwise executed at class scope, we need to set an
            # 'attr' into the settings if one isn't already in there
            if kwargs.get('attr') is None:
                kwargs['attr'] = wrapped.__name__

        return wrapped


def add_template_only_view(config: Configurator, pattern: str, name: str, template: str, view_args: t.Optional[dict]=None, route_args: t.Optional[dict]=None):
    """Adds a view which do not have a specific view function assgined.

    The view will render a template with the default template context.

    :param pattern: A path where the view is, e.g. ``/features``
    :param name: View name for ``route_url()``
    :param tempalte: A template to render
    :param view_args: kwargs passed to :py:meth:`pyramid.config.Configurator.add_view`
    :param route_args: kwargs passed to :py:meth:`pyramid.config.Configurator.add_view`
    """
    def _default_view(request):
        return {}

    config.add_route(name, pattern)
    config.add_view(view=_default_view, route_name=name, renderer=template)


def get_config_route(request: Request, config_key: str) -> str:
    """Route to a given URL from settings file."""
    settings = request.registry.settings

    try:
        return request.route_url(settings[config_key])
    except KeyError:
        return settings[config_key]



"""Transaction retry point support for command line applications and daemons.."""

# Standard Library
import logging
import threading
import typing as t
from functools import wraps

# Pyramid
import transaction
from transaction import TransactionManager


logger = logging.getLogger(__name__)


class NotRetryable(Exception):
    """Transaction retry mechanism not configured."""


class TransactionAlreadyInProcess(Exception):
    """ensure_transactionless() detected a dangling transactions."""


class CannotRetryAnymore(Exception):
    """We have reached the limit of transaction retry counts in @retryable."""


class TooDeepInTransactions(Exception):
    """@retryable function messed up with the transaction management"""


def ensure_transactionless(msg=None, transaction_manager=transaction.manager):
    """Make sure the current thread doesn't already have db transaction in process.

    :param transaction_manager: TransactionManager to check. Defaults to thread local transaction manager.
    """

    txn = transaction_manager._txn

    if txn:
        if not msg:
            msg = "Dangling transction open in transaction.manager. You should not start new one."

        transaction_thread = getattr(transaction.manager, "begin_thread", None)
        logger.fatal("Transaction state management error. Trying to start TX in thread %s. TX started in thread %s", threading.current_thread(), transaction_thread)

        # Destroy the transaction, so if this was a temporary failure in long running process, we don't lock the process up for the good
        txn.abort()

        raise TransactionAlreadyInProcess(msg)


def is_retryable(txn, error):
    """Check if this transaction is one caused by database conflict.

    These transactions should not be caught in catch all exception expressions.



    :param txn:
    :param error:
    :return:
    """
    # Emulate TransactionManager.is_retryable

    if txn is None:
        return False

    for dm in txn._resources:
        should_retry = getattr(dm, 'should_retry', None)
        if (should_retry is not None) and should_retry(error):
            return True


def retryable(tm: t.Optional[TransactionManager]=None, get_tm: t.Optional[t.Callable]=None):
    """Function decorator for§ SQL Serialized transaction conflict resolution through retries.

    You need to give either ``tm`` or ``get_tm`` argument.

    * New transaction is started when entering the decorated function

    * If there is already a transaction in progress when entering the decorated function raise an error

    * Commit when existing the decorated function

    * If the commit fails due to a SQL serialization conflict then try to rerun the decorated function max ``tm.retry_attempt_count`` times. Usually this is configured in TODO.

    Example:

    .. code-block:: python

        from websauna.system.model.retry import retryable

        def deposit_eth(web3: Web3, dbsession: Session, opid: UUID):

            @retryable(tm=dbsession.transaction_manager)
            def perform_tx():
                op = dbsession.query(CryptoOperation).get(opid)
                op.mark_performed()
                op.mark_broadcasted()
                # Transaction confirmation count updater will make sure we have enough blocks,
                # and then will call mark_completed()

            perform_tx()

    Example using class based transaction manager resolver:

    .. code-block:: python

        from websauna.system.model.retry import retryable

        class OperationQueueManager:

            def __init__(self, web3: Web3, dbsession: Session, asset_network_id, registry: Registry):
                assert isinstance(registry, Registry)
                assert isinstance(asset_network_id, UUID)
                self.web3 = web3
                self.dbsession = dbsession
                self.asset_network_id = asset_network_id
                self.registry = registry
                self.tm = self.dbsession.transaction_manager

            def _get_tm(*args, **kargs):
                self = args[0]
                return self.tm

            @retryable(get_tm=_get_tm)
            def get_waiting_operation_ids(self) -> List[Tuple[UUID, CryptoOperationType]]:
                wait_list = self.dbsession.query(CryptoOperation, CryptoOperation.id, CryptoOperation.state, CryptoOperation.operation_type).filter_by(network_id=self.asset_network_id, state=CryptoOperationState.waiting)

                # Flatten
                wait_list = [(o.id, o.operation_type) for o in wait_list]
                return wait_list

            def run_waiting_operations(self):
                # Performed inside TX retry boundary
                ops = self.get_waiting_operation_ids()

    Transaction manager needs ``retry_attempt_count`` attribute set by Websauna framework.

    :param tm: Transaction manager used to control the TX execution

    :param get_tm: Factory function that is called with ``\*args`` and ``\**kwargs`` to get the transaction manager
    """

    def _transaction_retry_wrapper(func):

        @wraps(func)
        def decorated_func(*args, **kwargs):

            global _retry_count

            if get_tm:
                manager = get_tm(*args, **kwargs)
            else:
                # Get how many attempts we want to do
                manager = tm

            assert manager, "No transaction manager available for retry"

            # Make sure we don't re-enter to transaction
            ensure_transactionless(transaction_manager=manager)
            retry_attempt_count = getattr(manager, "retry_attempt_count", None)
            if retry_attempt_count is None:
                raise NotRetryable("TransactionManager is not configured with default retry attempt count")

            # Run attempt loop
            latest_exc = None
            for num in range(retry_attempt_count):
                if num >= 1:
                    logger.info("Transaction attempt #%d for function %s", num + 1, func)

                txn = manager.begin()

                # Expose retry count for testing
                manager.latest_retry_count = num

                try:
                    val = func(*args, **kwargs)
                    try:
                        txn.commit()
                    except ValueError as ve:
                        # Means there was a nested transaction begin
                        raise TooDeepInTransactions("Looks like transaction.commit() failed - usually this means that the wrapped function {} begun its own transaction and ruined transaction state management".format(func)) from ve

                    return val
                except Exception as e:
                    if is_retryable(txn, e):
                        latest_exc = e
                        continue
                    else:
                        txn.abort()  # We could not commit
                        raise e

            raise CannotRetryAnymore("Out of transaction retry attempts, tried {} times".format(num + 1)) from latest_exc

        return decorated_func

    return _transaction_retry_wrapper



"""Default user object generator."""
# Standard Library
import typing as t
from datetime import timedelta

# Pyramid
from zope.interface import implementer

# SQLAlchemy
from sqlalchemy import func

# Websauna
from websauna.system.user.interfaces import IActivationModel
from websauna.system.user.interfaces import IPasswordHasher
from websauna.system.user.interfaces import IUser
from websauna.system.user.interfaces import IUserModel
from websauna.system.user.interfaces import IUserRegistry
from websauna.system.user.usermixin import GroupMixin
from websauna.system.user.usermixin import UserMixin
from websauna.system.user.utils import get_activation_model
from websauna.system.user.utils import get_user_class
from websauna.utils.time import now


@implementer(IUserRegistry)
class DefaultEmailBasedUserRegistry:
    """Default user backend which uses SQLAlchemy to store User models.

    Provides default user actions
    """

    def __init__(self, request):
        """Initialize DefaultEmailBasedUserRegistry.

        :param request:Pyramid request.
        """
        self.dbsession = request.dbsession
        self.registry = request.registry

    @property
    def User(self) -> t.Type[IUserModel]:
        """Currently configured User SQLAlchemy model.

        :return: Class implementing IUserModel.
        """
        return get_user_class(self.registry)

    @property
    def Activation(self) -> t.Type[IActivationModel]:
        """Currently configured User SQLAlchemy model.

        :return: Class implementing IActivationModel.
        """
        return get_activation_model(self.registry)

    def set_password(self, user: IUser, password: str):
        """Hash a password for persistent storage.

        Uses password hasher registered in :py:meth:`websauna.system.Initializer.configure_password`.

        :param user: User object.
        :param password: User password.
        """
        hasher = self.registry.getUtility(IPasswordHasher)
        hashed = hasher.hash_password(password)
        user.hashed_password = hashed

    def verify_password(self, user: IUser, password: str) -> bool:
        """Validate user password.

        Uses password hasher registered in :py:meth:`websauna.system.Initializer.configure_password`.

        :param user: User object.
        :param password: User password.
        :return: Boolean of the password verification.
        """
        if not user.hashed_password:
            # User password not set, always fail
            return False

        hasher = self.registry.getUtility(IPasswordHasher)
        return hasher.verify_password(user.hashed_password, password)

    def get_by_username(self, username: str) -> IUser:
        """Return the User with the given username.

        We always compare the lowercase version of User.username and username.

        :param username: Username to be used.
        :return: User object.
        """
        username = username.lower()
        user_class = self.User
        return self.dbsession.query(user_class).filter(func.lower(user_class.username) == username).first()

    def get_by_email(self, email: str) -> IUser:
        """Return the User with the given email.

        We always compare the lowercase version of User.email and email.

        :param email: Email to be used.
        :return: User object.
        """
        email = email.lower()
        user_class = self.User
        return self.dbsession.query(user_class).filter(func.lower(user_class.email) == email).first()

    def get_by_activation(self, activation: IActivationModel) -> IUser:
        """Return the User with the given activation.

        :param activation: Activation object..
        :return: User object.
        """
        user_class = self.User
        return self.dbsession.query(user_class).filter(user_class.activation_id == activation.id).first()

    def can_login(self, user: IUser) -> bool:
        """Verify if user is allowed do login.

        :param user: User object.
        :return: Boolean
        """
        return user.can_login()

    def get_groups(self, user: IUser) -> t.List[GroupMixin]:
        """Groups for a user.

        :param user: User object.
        :return: List of groups for this user.
        """
        return user.groups

    def create_password_reset_token(self, email: str) -> t.Optional[t.Tuple[object, str, int]]:
        """Sets password reset token for user.

        :param email: Email to be used.
        :return: [User, password reset token, token expiration in seconds]. ``None`` if user is disabled or is not email login based.
        """
        user = self.get_by_email(email)
        assert user, "Got password reset request for non-existing email".format(email)

        if not self.can_login(user):
            return None

        activation_token_expiry_seconds = int(self.registry.settings.get("websauna.activation_token_expiry_seconds", 24 * 3600))

        activation = self.Activation()
        activation.expires_at = now() + timedelta(seconds=activation_token_expiry_seconds)
        self.dbsession.add(activation)
        self.dbsession.flush()
        user.activation = activation

        assert user.activation.code, "Could not generate the password reset code"

        return user, activation.code, activation_token_expiry_seconds

    def create_email_activation_token(self, user: IUser) -> t.Tuple[str, int]:
        """Create activation token for the user to be used in the email

        :param user: User object.
        :return: Tuple (email activation code, expiration in seconds)
        """
        activation = self.Activation()
        activation_token_expiry_seconds = int(self.registry.settings.get("websauna.activation_token_expiry_seconds", 24 * 3600))
        activation.expires_at = now() + timedelta(seconds=activation_token_expiry_seconds)

        self.dbsession.add(activation)
        self.dbsession.flush()
        user.activation = activation
        return activation.code, activation_token_expiry_seconds

    def get_authenticated_user_by_username(self, username: str, password: str) -> t.Optional[UserMixin]:
        """Authenticate incoming user using username and password.

        :param username: Provided username.
        :param password: Provided password.
        :return: User instance of none if password does not match
        """
        user = self.get_by_username(username)
        if user and self.verify_password(user, password):
            return user
        return None

    def get_authenticated_user_by_email(self, email: str, password: str) -> t.Optional[UserMixin]:
        """Authenticate incoming user using email and password.

        :param email: Provided email.
        :param password: Provided password.
        :return: User instance of none if password does not match
        """
        user = self.get_by_email(email)
        if user and self.verify_password(user, password):
            return user
        return None

    def get_session_token(self, user: IUser):
        """Get marker string we use to store reference to this user in authenticated session.

        :param user: User object.
        :return: User id.
        """
        return user.id

    def get_user_by_session_token(self, token: str) -> UserMixin:
        """Resolve the authenticated user by a session token reference.

        :param token: Token to be used to return the user.
        :return: User object.
        """
        return self.dbsession.query(self.User).get(token)

    def get_user_by_password_reset_token(self, token: str) -> t.Optional[UserMixin]:
        """Get user by a password token issued earlier.

        Consume any activation token.

        :param token: Reset password token to be used to return the user.
        :return: User instance of none if token is not found.
        """
        activation = self.dbsession.query(self.Activation).filter(self.Activation.code == token).first()

        if activation:
            if activation.is_expired():
                return None
            user = self.get_by_activation(activation)
            return user
        return None

    def activate_user_by_email_token(self, token: str) -> t.Optional[UserMixin]:
        """Get user by a password token issued earlier.

        Consume any activation token.

        :param token: Password token to be used to return the user.
        :return: User instance of none if token is not found.
        """
        activation = self.dbsession.query(self.Activation).filter(self.Activation.code == token).first()

        if activation:
            if activation.is_expired():
                return None

            user = self.get_by_activation(activation)
            user.activated_at = now()
            self.dbsession.delete(activation)
            return user

        return None

    def reset_password(self, user: IUser, password: str):
        """Reset user password and clear all pending activation issues.

        :param user: User object,
        :param password: New password.
        """
        self.set_password(user, password)

        if not user.activated_at:
            user.activated_at = now()
        self.dbsession.delete(user.activation)

    def sign_up(self, registration_source: str, user_data: dict) -> UserMixin:
        """Sign up a new user through registration form.

        :param registration_source: Indication where the user came from.
        :param user_data: Payload with new user information.
        :return: Newly created User object.
        """
        password = user_data.pop("password", None)

        u = self.User(**user_data)

        if password:
            self.set_password(u, password)

        self.dbsession.add(u)
        self.dbsession.flush()

        # Generate default username (should not be exposed by default)
        u.username = u.generate_username()

        # Record how we created this record
        u.registration_source = registration_source

        self.dbsession.flush()
        return u



# Pyramid
import transaction

# Websauna
from websauna.system.user.models import User
from websauna.system.tests.utils import create_logged_in_user
from websauna.system.tests.utils import create_user


def test_pagination(web_server, browser, dbsession, init):
    with transaction.manager:
        create_logged_in_user(
            dbsession,
            init.config.registry,
            web_server,
            browser,
            admin=True
        )

        for index in range(1, 101):
            u = create_user(
                dbsession,
                init.config.registry,
                email="example{}@example.com".format(index)
            )
            dbsession.add(u)

    # quick check total users
    assert dbsession.query(User).count() == 101

    b = browser
    b.visit(web_server + "/admin/models/user/listing")

    # pagination should show correct number of total
    assert b.is_text_present("Total 101 items")
    assert b.is_text_present("Page #1 (1-20 of 101)")

    # page should show 20 rows (default size)
    assert len(b.find_by_css("tr.crud-row")) == 20

    # first email should be last created
    assert b.find_by_css("td.crud-column-email").first.text == "example100@example.com"

    # pager should show 2 buttons, first 2 are disabled
    assert len(b.find_by_css(".pager li")) == 4
    assert len(b.find_by_css(".pager li.disabled")) == 2

    # click to next and repeat the above tests
    b.find_by_css(".pager li")[2].click()
    assert b.is_text_present("Total 101 items")
    assert b.is_text_present("Page #2 (21-40 of 101)")
    assert len(b.find_by_css("tr.crud-row")) == 20
    assert b.find_by_css("td.crud-column-email").first.text == "example80@example.com"
    assert len(b.find_by_css(".pager li")) == 4
    assert len(b.find_by_css(".pager li.disabled")) == 0



"""Deform throttling support."""
# Standard Library
import logging
import typing as t

# Pyramid
import colander as c
from pyramid import httpexceptions

# Websauna
from websauna.system.core.redis import get_redis
from websauna.system.http import Request

from . import rollingwindow


logger = logging.getLogger(__name__)


def create_throttle_validator(name: str, max_actions_in_time_window: int, time_window_in_seconds: int=3600):
    """Creates a Colander form validator which prevents form submissions exceed certain rate.

    The benefit of using validator instead of :func:`throttle_view` decorator is that we can give a nice
    Colander form error message instead of an error page.

    Form submissions are throttled system wide. This prevents abuse of the system by flooding it with requests.

    A logging warning is issued if the rate is exceeded. The user is greeted with an error message telling the submission is not possible at the moment.

    Example::

        from tomb_routes import simple_route

        from websauna.system.form.throttle import create_throttle_validator

        from myapp import schemas

        @simple_route("/login", route_name="login", renderer="login/login.html", append_slash=False)
        def login(request):

            # Read allowed email login rate from the config file
            email_login_rate = int(request.registry.settings.get("trees.email_login_rate", 50))

            # Create a Colander schema instance with rate limit validator
            email_schema = schemas.LoginWithEmail(validator=create_throttle_validator("email_login", email_login_rate)).bind(request=request)

    :param name: Identify this throttler. Used as a Redis key.

    :param max_actions_in_time_window: Number of allowed actions per window

    :param time_window_in_seconds: Time in window in seconds. Default one hour, 3600 seconds.

    :return: Function to be passed to ``validator`` Colander schema construction parameter.
    """

    @c.deferred
    def throttle_validator(node, kw):
        """Protect invite functionality from flood attacks."""
        request = kw["request"]

        limit = max_actions_in_time_window

        def inner(node, value):
            # Check we don't have many invites going out
            if rollingwindow.check(request.registry, "throttle_" + name, window=time_window_in_seconds, limit=limit):

                # Alert devops through Sentry
                logger.warn("Excessive form submissions on %s", name)

                # Tell users slow down
                raise c.Invalid(node, 'Too many form submissions at the moment. Please try again later.')

        return inner

    return throttle_validator


def _read_throttle_settings(settings, setting):

    setting_value = settings.get(setting)
    if not setting_value:
        raise RuntimeError("Cannot read setting: {}".format(setting))

    parts = setting_value.split("/")
    if len(parts) != 2:
        raise RuntimeError("Bad throttle setting format: {}".format(setting_value))

    try:
        window = int(parts[0])
        limit = int(parts[1])
    except ValueError:
        raise RuntimeError("Could not parse: {}".format(setting_value))

    return limit, window


def throttled_view(
        rolling_window_id: t.Optional[str]=None,
        time_window_in_seconds: int=3600,
        limit: int=50,
        setting: t.Optional[str]=None):
    """Decorate a view to protect denial-of-service attacks using throttling.

    If the global throttling limit is exceeded the client gets HTTP 429 error.

    Internally we use :py:mod:`websauna.system.form.rollingwindow` hit counting implementation.

    Example that allows 30 page loads per hour:

    .. code-block:: python

        @view_config(
            name="new-phone-number",
            renderer="wallet/new_phone_number.html",
            decorator=throttled_view(limit=30, time_window_in_seconds=3600))
        def new_phone_number(request):
            # ... code goes here ...

    You can also configure the limit in a INI settings file.

    Example ``production.ini``:

    .. code-block:: ini

        magiclogin.email_throttle = 50/3600

    .. code-block:: python

        @view_config(
            name="email-login",
            renderer="wallet/new_phone_number.html",
            decorator=throttled_view(setting="magiclogin.email_throttle")
        def new_phone_number(request):
            # ... code goes here ...

    :param rolling_window_id: The Redis key name used to store throttle state. If not given a key derived from view name is used.

    :param time_window_in_seconds: Sliding time window for counting hits

    :param limit: Allowed number of hits in the given time window

    :param setting: Read throttle information from a INI settings. In this case string must be in format limit/time_window_seconds. Example: 50/3600. This overrides any given time and limit argument.

    :raise: :py:class:`pyramid.httpexceptions.HTTPTooManyRequests` if the endpoint gets hammered too much
    """
    # http://docs.pylonsproject.org/projects/pyramid_cookbook/en/latest/views/chaining_decorators.html
    def outer(view_callable: t.Callable):

        name = view_callable.__name__

        if not rolling_window_id:
            key_name = "throttle_{key}".format(key=name)
        else:
            key_name = "throttle_{key}".format(key=rolling_window_id)

        def inner(context, request):

            if setting:
                hits, window = _read_throttle_settings(request.registry.settings, setting)
            else:
                window = time_window_in_seconds
                hits = limit

            if rollingwindow.check(request.registry, key_name, window=window, limit=hits):
                raise httpexceptions.HTTPTooManyRequests("Too many requests against {}".format(name))

            return view_callable(context, request)

        return inner

    return outer


def clear_throttle(request: Request, key_name: str):
    """Clear the throttling status.

    Example:

    .. code-block:: python

        clear_throttle(request, "new-phone-number")

    """
    redis = get_redis(request)
    redis.delete("throttle_{}".format(key_name))



# Pyramid
from pyramid.events import subscriber

# Websauna
from websauna.system.admin import menu
from websauna.system.admin.events import AdminConstruction
from websauna.system.admin.modeladmin import ModelAdminRoot
from websauna.system.core.traversal import Resource


@subscriber(AdminConstruction)
def contribute_model_admin(event):
    """Add model menus to the admin user interface."""

    admin = event.admin
    request = event.admin.request

    model_admin_root = ModelAdminRoot(request)
    Resource.make_lineage(admin, model_admin_root, "models")
    admin.children["models"] = model_admin_root

    # Create a model entries to menu
    data_menu = admin.get_admin_menu().get_entry("admin-menu-data").submenu
    for id, model_admin in model_admin_root.items():
        entry = menu.TraverseEntry("admin-menu-data-{}".format(id), label=model_admin.title, resource=model_admin, name="listing")
        data_menu.add_entry(entry)



"""Simple mechanism to register admin menu entries."""
# Standard Library
import typing as t
from collections import OrderedDict

# Pyramid
from pyramid.renderers import render
from pyramid.request import Request

# Websauna
from websauna.system.core.traversal import Resource


class Menu:
    """Menu is a collection of items in a Bootstrap pull-down menu.

    It allow mechanism for third party packages to register their own entries.
    """

    template = "admin/menu/menu.html"

    def __init__(self):
        self.entries = OrderedDict()

    # TODO: How to define entry type here to be Entry (circular)
    def add_entry(self, entry):
        self.entries[entry.id] = entry

    def has_items(self, request: Request) -> bool:
        """Has this menu any entries to draw."""
        return any(entry.is_enabled(request) for entry in self.entries.values())

    def get_entries(self) -> t.List:
        """Get Entry objects to be rendered in this menu.

        Sort return by natural name order.
        """
        return sorted(self.entries.values(), key=lambda e: e.label)

    def get_entry(self, id):
        """Get any of registered menu entries by its id."""
        return self.entries[id]


class Entry:
    """Option is one choice in a menu.

    It can be a simple link, submenu or a custom rendered template.

    Adding conditional entry with a permission check::

        admin = Admin.get_admin(self.config.registry)
        entry = menu.RouteEntry("admin-notebook", label="Shell", icon="fa-terminal", route_name="admin_shell", condition=lambda request:request.has_permission('shell'))
        admin.get_root_menu().add_entry(entry)
    """

    template = "admin/menu/entry.html"

    caret = "fa fa-caret-down"

    def __init__(self, id: str, label: str, icon: str=None, caret: str=None, css_class: str=None, template: str=None, submenu: Menu=None, condition: t.Callable=None, link: t.Callable=None, extra: dict=None):
        """
        :param id: Machine id and CSS id for this menu entry
        :param label: Human-readable label of the menu item
        :param css_class: CSS class for the main entry HTML element
        :param icon: CSS class for FontAwesome <i> icon. Example: ``fa-wrench``.
        :param caret: CSS class for submenu caret. Example: ```fa fa-caret-right``. Default: ``fa fa-caret-down``
        :param template: Jinja 2 template definition used to render this entry. Defaults to ``admin/menu_options.html``
        :param submenu: If this entry is a submenu, then this points to the menu object.
        :param condition: callable(entry, request) which returns True or False should the menu entry be visible
        :param link: callable(entry, request) which returns string to a where this menu entry should go
        :param extra: Dictionary of extra parameters which can be accessed later on in templates and functions
        """
        assert id
        self.id = id
        self.label = label
        self.icon = icon
        self.submenu = submenu
        self.css_class = css_class
        self.condition = condition

        if caret:
            self.caret = caret

        if template:
            self.template = template

        self.extra = extra or {}

    def is_enabled(self, request) -> bool:
        """Return True if this entry is visible."""

        # Is this a submenu without entries
        if self.submenu:
            if not self.submenu.has_items(request):
                return False

        # The condition for this entry valuates to False
        if self.condition is not None:
            return self.condition(self, request)

        return True

    def get_link(self, request: Request) -> str:
        """Get the link target where this menu entry jumps to."""

        if self.link:
            return self.link(self, request)

        raise NotImplementedError()

    def render(self, request: Request):
        """Render this item to HTML.

        :return: Rendered HTML as a string
        """
        context = dict(entry=self)
        return render(self.template, context, request=request)


class RouteEntry(Entry):
    """Menu entry which has a Pyramid route as a link. """

    def __init__(self, id: str, label: str, route_name: str, **kwargs):
        """
        :param route_name: Link target name for ``request.route_url()``
        """
        assert route_name
        assert type(route_name) == str
        super(RouteEntry, self).__init__(id, label, **kwargs)
        self.route_name = route_name

    def get_link(self, request):
        return request.route_url(self.route_name)


class TraverseEntry(Entry):
    """Menu entry which has a Pyramid traversing context as a link.

    Please note that this works only for fixed resources which are generated on the application startup.

    Example::

        admin = Admin.get_admin(self.config.registry)
        entry = menu.TraverseEntry(id="admin-menu-phone-order", label="Create phone order", context=admin, name="phone-order", icon="fa-phone")
        admin.get_admin_menu().add_entry(entry)
    """

    def __init__(self, id: str, label: str, resource: Resource, name: str, **kwargs):
        """
        :param context: Any Pyramid resource object
        :param name: Traversable view name
        """
        assert name
        assert type(name) == str
        super(TraverseEntry, self).__init__(id, label, **kwargs)
        self.resource = resource
        self.name = name

    def get_link(self, request):
        return request.resource_url(self.resource, self.name)


class NavbarEntry(Entry):
    """Root entry for rendering horizontal navigation list menu."""

    template = "admin/menu/navbar.html"



"""Default template variables."""
# Standard Library
import datetime
import logging

# Pyramid
from pyramid.events import BeforeRender
# TODO: We expose this as a global for documentation. Migrate to non-global map in the future.
from pyramid.path import AssetResolver  # noQA
from pyramid.path import DottedNameResolver

# Websauna
from websauna.utils.time import now


logger = logging.getLogger(__name__)


_template_variables = {}


def var(name):
    """Decorator to mark template variables for documentation."""

    def _inner(func):
        _template_variables[name] = func
        return func

    return _inner


@var("site_name")
def site_name(request, registry, settings):
    """Expose website name from ``websauna.site_name`` config variable to templates.

    Example:

    .. code-block:: html+jinja

        <div class="jumbotron text-center">
            <h1>{{ site_name }}</h1>
            <p class="lead text-center">
                {{ site_tag_line }}
            </p>
        </div>

    """
    return settings["websauna.site_name"]


@var("site_title")
def site_title(request, registry, settings):
    """Expose website name from ``websauna.site_title`` config variable to templates.

    This is the default ``<title>`` tag.

    Example:

    .. code-block:: html+jinja

        <meta>
            <title>My page - {{ site_title }}</title>
        </meta>

    """
    # Use .get() for BBB
    return settings.get("websauna.site_title", "")


@var("site_url")
def site_url(request, registry, settings):
    """Expose website URL from ``websauna.site_url`` config variable to templates.

    .. note ::

        You should not use this variable in web page templates. This variable is intended for cases where one needs templating without running a web server.

    The correct way to get the home URL of your website is:

    .. code-block:: html+jinja

        <a href="{{ request.route_url('home') }}">Home</a>

    """
    return settings["websauna.site_url"]


@var("site_author")
def site_author(request, registry, settings):
    """Expose website URL from ``websauna.site_author`` config variable to templates.

    This is used in footer to display the site owner.
    """
    return settings["websauna.site_author"]


@var("site_tag_line")
def site_tag_line(request, registry, settings):
    """Expose website URL from ``websauna.site_tag_line`` config variable to templates.

    This is used on the default front page to catch the attention of audience.
    """
    return settings["websauna.site_tag_line"]


@var("site_email_prefix")
def site_email_prefix(request, registry, settings):
    """Expose website URL from ``websauna.site_email_prefix`` config variable to templates.

    This is used as the subject prefix in outgoing email. E.g. if the value is ``SuperSite`` you'll email subjects::

        [SuperSite] Welcome to www.supersite.com

    """
    return settings["websauna.site_email_prefix"]


@var("site_time_zone")
def site_time_zone(request, registry, settings):
    """Expose website URL from ``websauna.site_time_zone`` config variable to templates.

    By best practices, all dates and times should be stored in the database using :term:`UTC` time. This setting allows quickly convert dates and times to your local time.

    Example:

    .. code-block:: html+jinja

        <p>
            <strong>Bar opens</strong>:
            {{ opening_at|friendly_time(timezone=site_time_zone) }}
        </p>

    Default value is ``UTC``.

    See `timezone abbreviation list <https://en.wikipedia.org/wiki/List_of_time_zone_abbreviations>`_.
    """
    return settings.get("websauna.site_timezone", "UTC")


@var("js_in_head")
def js_in_head(request, registry, settings):
    """Should ``<script>`` tags be placed in ``<head>`` or end of ``<body>``.

    """
    on_demand_resource_renderer = getattr(request, "on_demand_resource_renderer", None)
    if on_demand_resource_renderer and on_demand_resource_renderer.is_js_in_head(request):
        return True
    else:
        return False


@var("on_demand_resource_renderer")
def on_demand_resource_renderer(request, registry, settings):
    """Active instance of :py:class:`websauna.system.core.render.OnDemandResourceRenderer` managing dynamic CSS and JS. May be None."""
    on_demand_resource_renderer = getattr(request, "on_demand_resource_renderer", None)
    return on_demand_resource_renderer


@var("now")
def _now(request, registry, settings):
    """Get the current time as :term:`UTC`.

    :py:func:`websauna.utils.time.now` function.

    Example:

    .. code-block: html+jinja

      <footer class="footer">
        <div class="container-fluid">
          <p class="copyright pull-right">
            &copy; {{ now().year }} {{ site_author }}
          </p>
        </div>
      </footer>

    """
    return now


@var("timedelta")
def _timedelta(request, registry, settings):
    """Expose Python timedelta class to templates, so you can do relative time calculations.

    For more information see :py:class:`datetime.timedelta`.

    Example:

    .. code-block: html+jinja

        <div class="panel panel-default panel-admin">
          <div class="panel-body">
            <h2>
                Current power generation
                (as generated {{ (now() - timedelta(days=14))|arrow_format("YYYY-MM-DD")}}*)
            </h2>

            <div id="super-nice-graph"></div>

            <p>* Incoming data delayed two weeks</p>

          </div>
        </div>

    """
    return datetime.timedelta


@var("debug")
def debug(request, registry, settings):
    """Invoke pdb breakpoint from a template.

    Example:

    .. code-block:: html+jinja

        <h1>{{ site_name }}</h1>

        {{ debug() }}

    This will invoke function from :ref:`websauna.template_debugger` setting. The debugger is turned on only on :ref:`development.ini`. If there is no debugger configured, a warning is given.
    """

    def _dummy():
        logger.warn("{{ debug() }} invoked, but websauna.template_debugger not set")
        return ""

    template_debugger = settings.get("websauna.template_debugger")
    if not template_debugger:
        debugger = _dummy
    else:
        r = DottedNameResolver()
        debugger = r.resolve(template_debugger)

    def _inner():
        debugger()
        return ""

    return _inner


def includeme(config):

    def on_before_render(event):
        # Augment Pyramid template renderers with these extra variables and deal with JS placement

        request = event["request"]

        for name, func in _template_variables.items():
            event[name] = func(request, request.registry, request.registry.settings)

    config.add_subscriber(on_before_render, BeforeRender)
